# coding: utf-8
# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.
"""
Functions for creating supercells for NEB calculations
"""
import logging
from typing import List, Tuple, Union, Optional

import numpy as np

# from ase.build import find_optimal_cell_shape, get_deviation_from_optimal_cell_shape
# from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.core import PeriodicSite, Structure
from pymatgen.transformations.advanced_transformations import CubicSupercellTransformation

__author__ = "Jimmy Shen"
__copyright__ = "Copyright 2019, The Materials Project"
__maintainer__ = "Jimmy Shen"
__email__ = "jmmshn@lbl.gov"
__date__ = "Feb 11, 2021"

logger = logging.getLogger(__name__)

# Helper functions for MigraionHop.get_sc_struture


def get_sc_fromstruct(
    base_struct: Structure,
    min_atoms: int = 80,
    max_atoms: int = 240,
    min_length: float = 10.0,
) -> List[List[int]]:
    """
    Generate the best supercell from a unitcell.
    The CubicSupercellTransformation from PMG is much faster but don't iterate over as many
    supercell configurations so it's less able to find the best configuration in a give cell size.
    We try the PMG's cubic supercell transformation with a cap on the number of atoms (max_atoms).
    The min_length is decreased by 10% (geometrically) until a supercell can be constructed.

    Args:
        base_struct: structure of the unit cell
        max_atoms: Maximum number of atoms allowed in the supercell.
        min_atoms: Minimum number of atoms allowed in the supercell.
        min_length: Minimum length of the smallest supercell lattice vector.

    Returns:
        struc_sc: Supercell that is as close to cubic as possible
    """
    m_len = min_length
    struct_sc = None
    while struct_sc is None:
        struct_sc = _get_sc_from_struct_pmg(base_struct, min_atoms, max_atoms, m_len)
        m_len *= 0.9
    return struct_sc


def _get_sc_from_struct_pmg(
    base_struct: Structure,
    min_atoms: int = 80,
    max_atoms: int = 240,
    min_length: float = 10.0,
) -> Optional[List[List[int]]]:
    """
    Generate the best supercell from a unitcell using the pymatgen CubicSupercellTransformation

    Args:
        base_struct: structure of the unit cell
        max_atoms: Maximum number of atoms allowed in the supercell.
        min_atoms: Minimum number of atoms allowed in the supercell.
        min_length: Minimum length of the smallest supercell lattice vector.

    Returns:
        3x3 matrix: supercell matrix

    """
    cst = CubicSupercellTransformation(min_atoms=min_atoms, max_atoms=max_atoms, min_length=min_length)

    try:
        cst.apply_transformation(base_struct)
    except BaseException:
        return None
    return cst.transformation_matrix


# Something is broken with how ASE generates supercells
# now many calls to `find_optimal_cell_shape` results in
# `Failed to find a transformation matrix`
# So remove this funtionality now in favour of decreasing the max_length
#
# def _get_sc_from_struct_ase(
#     base_struct: Structure, min_atoms: int=80, max_atoms: int=240
# ) -> List[List[int]]:
#     """generate the best supercell from a unitcell using ASE's method, is slower but more exhaustive
#
#     Args:
#         base_struct (pymatgen.Structure): unit cell
#         min_atoms (int, optional): Minimum number of atoms in the desired supercell. Defaults to 80.
#
#     Returns:
#         3x3 matrix: supercell matrix
#     """
#
#     num_cells_min = int(np.ceil(min_atoms / base_struct.num_sites))
#     num_cells_max = int(np.ceil(max_atoms / base_struct.num_sites))
#
#     atoms = AseAtomsAdaptor().get_atoms(base_struct)
#     res = []
#     for icell in range(num_cells_min, num_cells_max):
#         if icell % 2 != 0 and icell % 3 != 0:
#             continue  # cells with many factors are more likely to be square
#         logger.info(f"Getting cell shape {icell} x unit cell")
#         sc_mat = find_optimal_cell_shape(atoms.cell, icell, "sc")
#         if sc_mat is None:
#             continue
#         deviation = get_deviation_from_optimal_cell_shape(np.dot(sc_mat, atoms.cell))
#         dd = {"deviation": deviation, "P1": sc_mat}
#         # base_struct_sc.to('poscar', f'/Users/lik/Desktop/enpoints_{base_struct_sc.num_sites}.vasp')
#         res.append(dd)
#         if deviation < 0.3:
#             logger.debug("Found good cell")
#             return sc_mat
#
#     else:
#         best_case = min(res, key=lambda x: x["deviation"])
#     logger.warning(
#         f"Could not find case with deviation from cubic was less than 0.3 using the ASE cubic supercell finder \
#         \nThe best one had a deviation of {best_case['deviation']}"
#     )
#     return best_case["P1"]


def get_start_end_structures(
    isite: PeriodicSite,
    esite: PeriodicSite,
    base_struct: Structure,
    sc_mat: List[List[Union[int, float]]],
    vac_mode: bool,
    debug: bool = False,
) -> Tuple[Structure, Structure, Structure]:
    """
    Obtain the starting and terminating structures in a supercell for NEB calculations.

    Args:
        hop: object presenting the migration event
        base_struct: unit cell representation of the structure
        sc_mat: supercell transformation to create the simulation cell for the NEB calc

    Returns:
        initial structure, final structure, empty structure all in the supercell
    """

    def remove_site_at_pos(structure: Structure, site: PeriodicSite):
        new_struct_sites = []
        for isite in structure:
            if not vac_mode or (isite.distance(site) <= 1e-8):
                continue
            new_struct_sites.append(isite)
        return Structure.from_sites(new_struct_sites)

    base_sc = base_struct.copy() * sc_mat

    start_struct = base_struct.copy() * sc_mat
    end_struct = base_struct.copy() * sc_mat

    sc_mat_inv = np.linalg.inv(sc_mat)

    if not vac_mode:
        # insertion the endpoints
        start_struct.insert(
            0,
            esite.species_string,
            np.dot(isite.frac_coords, sc_mat_inv),
            properties={"magmom": 0},
        )
        end_struct.insert(
            0,
            esite.species_string,
            np.dot(esite.frac_coords, sc_mat_inv),
            properties={"magmom": 0},
        )
    else:
        # remove the other endpoint
        ipos_sc = np.dot(isite.frac_coords, sc_mat_inv)
        epos_sc = np.dot(esite.frac_coords, sc_mat_inv)
        if debug:
            icart = base_sc.lattice.get_cartesian_coords(ipos_sc)
            ecart = base_sc.lattice.get_cartesian_coords(epos_sc)
            assert abs(np.linalg.norm(icart - ecart) - np.linalg.norm(isite.coords - esite.coords)) < 1e-5
        i_ref_ = PeriodicSite(species=esite.species_string, coords=ipos_sc, lattice=base_sc.lattice)
        e_ref_ = PeriodicSite(species=esite.species_string, coords=epos_sc, lattice=base_sc.lattice)
        start_struct = remove_site_at_pos(start_struct, e_ref_)
        end_struct = remove_site_at_pos(end_struct, i_ref_)
    return start_struct, end_struct, base_sc



# coding: utf-8
# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.
from monty.serialization import loadfn
from pymatgen.entries.computed_entries import ComputedStructureEntry

from pymatgen.analysis.structure_matcher import StructureMatcher
import unittest
from pymatgen.core import Structure
import numpy as np
import os

from pymatgen.analysis.diffusion.utils.parse_entries import (
    get_inserted_on_base,
    process_entries,
    get_insertion_energy,
    get_sym_migration_ion_sites,
    _filter_and_merge,
)

dir_path = os.path.dirname(os.path.realpath(__file__)) + "/test_files"

__author__ = "Jimmy Shen"
__version__ = "1.0"
__date__ = "April 10, 2019"


class ParseEntriesTest(unittest.TestCase):
    def setUp(self):
        d = loadfn(f"{dir_path}/parse_entry_test_vars.json")
        struct_uc = d["struct_uc"]
        self.li_ent = d["li_ent"]
        e_uc = 100
        self.base = ComputedStructureEntry(structure=struct_uc, energy=e_uc)

        sc = struct_uc * [2, 2, 2]
        sc.insert(0, "Li", [0.125, 0.125, 0.25])
        self.inserted_1Li1 = ComputedStructureEntry(structure=sc, energy=e_uc * 8 + 3)

        sc = struct_uc * [2, 2, 2]
        sc.insert(0, "Li", [0.375, 0.375, 0.25])
        self.inserted_1Li2 = ComputedStructureEntry(structure=sc, energy=e_uc * 8 + 5)

        sc = struct_uc * [2, 2, 2]
        sc.insert(0, "Li", [0.125, 0.125, 0.25])
        sc.insert(0, "Li", [0.375, 0.375, 0.25])
        self.inserted_2Li = ComputedStructureEntry(structure=sc, energy=e_uc * 8 + 4)

        self.sm = StructureMatcher(ignored_species=["Li"], primitive_cell=False)

        self.struct_inserted_1Li1 = get_inserted_on_base(self.base, self.inserted_1Li1, self.li_ent, self.sm)
        self.struct_inserted_1Li2 = get_inserted_on_base(self.base, self.inserted_1Li2, self.li_ent, self.sm)
        self.struct_inserted_2Li = get_inserted_on_base(self.base, self.inserted_2Li, self.li_ent, self.sm)

    def _is_valid_inserted_ent(self, mapped_struct):
        res = True
        for isite in mapped_struct.sites:
            if isite.species_string == "Li":
                tt = isite.frac_coords - np.floor(isite.frac_coords)
                if list(tt) in [[0.75, 0.75, 0.5], [0.25, 0.25, 0.5]]:
                    res = True
                else:
                    return False
        return res

    def test_get_inserted_on_base(self):
        mapped_struct = get_inserted_on_base(self.base, self.inserted_1Li1, self.li_ent, self.sm)
        self.assertTrue(self._is_valid_inserted_ent(mapped_struct))
        self.assertEqual(mapped_struct[0].properties["insertion_energy"], 5.0)
        mapped_struct = get_inserted_on_base(self.base, self.inserted_1Li2, self.li_ent, self.sm)
        self.assertTrue(self._is_valid_inserted_ent(mapped_struct))
        self.assertEqual(mapped_struct[0].properties["insertion_energy"], 7.0)
        mapped_struct = get_inserted_on_base(self.base, self.inserted_2Li, self.li_ent, self.sm)
        self.assertTrue(self._is_valid_inserted_ent(mapped_struct))
        self.assertEqual(mapped_struct[0].properties["insertion_energy"], 4.0)

    def test_process_ents(self):
        base_2_ent = ComputedStructureEntry(
            structure=self.base.structure * [[1, 1, 0], [1, -1, 0], [0, 0, 2]],
            energy=self.base.energy * 4,
        )
        res = process_entries(
            [base_2_ent, self.base],
            [self.inserted_2Li],
            migrating_ion_entry=self.li_ent,
        )
        for itr_group in res:
            for i_insert_site in itr_group["inserted"]:
                if i_insert_site.species_string == "Li":
                    self.assertEqual(i_insert_site.properties["insertion_energy"], 4)

    def test_filter_and_merge(self):
        combined_struct = Structure.from_sites(
            self.struct_inserted_1Li1.sites + self.struct_inserted_1Li2.sites + self.struct_inserted_2Li.sites
        )
        filtered_struct = _filter_and_merge(combined_struct)
        for i_insert_site in filtered_struct:
            if i_insert_site.species_string == "Li":
                self.assertIn(i_insert_site.properties["insertion_energy"], {4.5, 5.5})

    def test_get_insertion_energy(self):
        insert_energy = get_insertion_energy(self.base, self.inserted_1Li1, self.li_ent)
        basex2_ = ComputedStructureEntry(structure=self.base.structure * [1, 1, 2], energy=self.base.energy * 2)
        insert_energyx2 = get_insertion_energy(basex2_, self.inserted_1Li1, self.li_ent)
        self.assertAlmostEqual(insert_energyx2, insert_energy)
        self.assertAlmostEqual(insert_energy, 5)  # 3 + 2 where 2 is from the Li energy
        insert_energy = get_insertion_energy(self.base, self.inserted_2Li, self.li_ent)
        self.assertAlmostEqual(insert_energy, 4)

    def test_get_all_sym_sites(self):
        struct11 = get_sym_migration_ion_sites(self.base.structure, self.inserted_1Li1.structure, migrating_ion="Li")
        self.assertEquals(struct11.composition["Li"], 4)
        struct12 = get_sym_migration_ion_sites(self.base.structure, self.inserted_1Li2.structure, migrating_ion="Li")
        self.assertEquals(struct12.composition["Li"], 4)
        struct22 = get_sym_migration_ion_sites(self.base.structure, self.inserted_2Li.structure, migrating_ion="Li")
        self.assertEquals(struct22.composition["Li"], 8)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.

"""
 Algorithms for diffusion pathway analysis
"""

from collections import Counter

import numpy as np
from scipy.cluster.hierarchy import fcluster, linkage
from scipy.spatial.distance import squareform


class ProbabilityDensityAnalysis:
    r"""
    Compute the time-averaged probability density distribution of selected
    species on a "uniform" (in terms of fractional coordinates) 3-D grid.
    Note that \int_{\Omega}d^3rP(r) = 1

    If you use this class, please consider citing the following paper:

    Zhu, Z.; Chu, I.-H.; Deng, Z. and Ong, S. P. "Role of Na+ Interstitials and
    Dopants in Enhancing the Na+ Conductivity of the Cubic Na3PS4 Superionic
    Conductor". Chem. Mater. (2015), 27, pp 8318–8325.
    """

    def __init__(self, structure, trajectories, interval=0.5, species=("Li", "Na")):
        """
        Initialization.
        Args:
            structure (Structure): crystal structure
            trajectories (numpy array): ionic trajectories of the structure
                from MD simulations. It should be (1) stored as 3-D array [
                Ntimesteps, Nions, 3] where 3 refers to a,b,c components;
                (2) in fractional coordinates.
            interval(float): the interval between two nearest grid points
                (in Angstrom)
            species(list of str): list of species that are of interest
        """

        # initial settings
        trajectories = np.array(trajectories)

        # All fractional coordinates are between 0 and 1.
        trajectories -= np.floor(trajectories)
        assert np.all(trajectories >= 0) and np.all(trajectories <= 1)

        indices = [j for j, site in enumerate(structure) if site.specie.symbol in species]
        lattice = structure.lattice
        frac_interval = [interval / l for l in lattice.abc]
        nsteps = len(trajectories)

        # generate the 3-D grid
        ra = np.arange(0.0, 1.0, frac_interval[0])
        rb = np.arange(0.0, 1.0, frac_interval[1])
        rc = np.arange(0.0, 1.0, frac_interval[2])
        lens = [len(ra), len(rb), len(rc)]
        ngrid = lens[0] * lens[1] * lens[2]

        agrid = ra[:, None] * np.array([1, 0, 0])[None, :]
        bgrid = rb[:, None] * np.array([0, 1, 0])[None, :]
        cgrid = rc[:, None] * np.array([0, 0, 1])[None, :]

        grid = agrid[:, None, None] + bgrid[None, :, None] + cgrid[None, None, :]

        # Calculate time-averaged probability density function distribution Pr
        count = Counter()
        Pr = np.zeros(ngrid, dtype=np.double)

        for it in range(nsteps):
            fcoords = trajectories[it][indices, :]
            for fcoord in fcoords:
                # for each atom at time t, find the nearest grid point from
                # the 8 points that surround the atom
                corner_i = [int(c / d) for c, d in zip(fcoord, frac_interval)]
                next_i = np.zeros_like(corner_i, dtype=int)

                # consider PBC
                for i in range(3):
                    next_i[i] = corner_i[i] + 1 if corner_i[i] < lens[i] - 1 else 0

                agrid = np.array([corner_i[0], next_i[0]])[:, None] * np.array([1, 0, 0])[None, :]
                bgrid = np.array([corner_i[1], next_i[1]])[:, None] * np.array([0, 1, 0])[None, :]
                cgrid = np.array([corner_i[2], next_i[2]])[:, None] * np.array([0, 0, 1])[None, :]

                grid_indices = agrid[:, None, None] + bgrid[None, :, None] + cgrid[None, None, :]
                grid_indices = grid_indices.reshape(8, 3)

                mini_grid = [grid[indx[0], indx[1], indx[2]] for indx in grid_indices]
                dist_matrix = lattice.get_all_distances(mini_grid, fcoord)
                indx = np.where(dist_matrix == np.min(dist_matrix, axis=0)[None, :])[0][0]

                # 3-index label mapping to single index
                min_indx = (
                    grid_indices[indx][0] * len(rb) * len(rc) + grid_indices[indx][1] * len(rc) + grid_indices[indx][2]
                )

                # make sure the index does not go out of bound.
                assert 0 <= min_indx < ngrid

                count.update([min_indx])

        for i, n in count.most_common(ngrid):
            Pr[i] = float(n) / nsteps / len(indices) / lattice.volume * ngrid

        Pr = Pr.reshape(lens[0], lens[1], lens[2])

        self.structure = structure
        self.trajectories = trajectories
        self.interval = interval
        self.lens = lens
        self.Pr = Pr
        self.species = species
        self.stable_sites = None

    @classmethod
    def from_diffusion_analyzer(cls, diffusion_analyzer, interval=0.5, species=("Li", "Na")):
        """
        Create a ProbabilityDensityAnalysis from a diffusion_analyzer object.

        Args:
            diffusion_analyzer (DiffusionAnalyzer): A
                pymatgen.analysis.diffusion_analyzer.DiffusionAnalyzer object
            interval(float): the interval between two nearest grid points (in
                Angstrom)
            species(list of str): list of species that are of interest
        """
        structure = diffusion_analyzer.structure
        trajectories = []

        for i, s in enumerate(diffusion_analyzer.get_drift_corrected_structures()):
            trajectories.append(s.frac_coords)

        trajectories = np.array(trajectories)

        return ProbabilityDensityAnalysis(structure, trajectories, interval=interval, species=species)

    def generate_stable_sites(self, p_ratio=0.25, d_cutoff=1.0):
        """
        Obtain a set of low-energy sites from probability density function with
        given probability threshold 'p_ratio'. The set of grid points with
        probability density higher than the threshold will further be clustered
        using hierachical clustering method, with no two clusters closer than the
        given distance cutoff. Note that the low-energy sites may converge more
        slowly in fast conductors (more shallow energy landscape) than in the slow
        conductors.

        Args:
            p_ratio (float): Probability threshold above which a grid point is
                considered as a low-energy site.
            d_cutoff (float): Distance cutoff used in hierachical clustering.

        Notes:
            The set of stable sites is stored in the `stable_sites` attribute
            as a Nx3 numpy array.
        """

        # Set of grid points with high probability density.
        grid_fcoords = []
        indices = np.where(self.Pr > self.Pr.max() * p_ratio)
        lattice = self.structure.lattice

        for (x, y, z) in zip(indices[0], indices[1], indices[2]):
            grid_fcoords.append([x / self.lens[0], y / self.lens[1], z / self.lens[2]])

        grid_fcoords = np.array(grid_fcoords)
        dist_matrix = np.array(lattice.get_all_distances(grid_fcoords, grid_fcoords))
        np.fill_diagonal(dist_matrix, 0)

        # Compressed distance matrix
        condensed_m = squareform((dist_matrix + dist_matrix.T) / 2.0)

        # Linkage function for hierachical clustering.
        z = linkage(condensed_m, method="single", metric="euclidean")
        cluster_indices = fcluster(z, t=d_cutoff, criterion="distance")

        # The low-energy sites must accommodate all the existing mobile ions.
        nions = len([e for e in self.structure if e.specie.symbol in self.species])
        nc = len(set(cluster_indices))

        if nc < nions:
            raise ValueError(
                "The number of clusters ({}) is smaller than that of "
                "mobile ions ({})! Please try to decrease either "
                "'p_ratio' or 'd_cut' values!".format(nc, nions)
            )

        # For each low-energy site (cluster centroid), its coordinates are obtained
        # by averaging over all the associated grid points within that cluster.
        stable_sites = []

        for i in set(cluster_indices):
            indices = np.where(cluster_indices == i)[0]

            if len(indices) == 1:
                stable_sites.append(grid_fcoords[indices[0]])
                continue

            # Consider periodic boundary condition
            members = grid_fcoords[indices] - grid_fcoords[indices[0]]
            members = np.where(members > 0.5, members - 1.0, members)
            members = np.where(members < -0.5, members + 1.0, members)
            members += grid_fcoords[indices[0]]

            stable_sites.append(np.mean(members, axis=0))

        self.stable_sites = np.array(stable_sites)

    def get_full_structure(self):
        """
        Generate the structure with the low-energy sites included. In the end, a
        pymatgen Structure object will be returned.
        """

        full_structure = self.structure.copy()
        for fcoord in self.stable_sites:
            full_structure.append("X", fcoord)

        return full_structure

    def to_chgcar(self, filename="CHGCAR.vasp"):
        """
        Save the probability density distribution in the format of CHGCAR,
        which can be visualized by VESTA.
        """

        count = 1
        VolinAu = self.structure.lattice.volume / 0.5291772083 ** 3
        symbols = self.structure.symbol_set
        natoms = [str(int(self.structure.composition[symbol])) for symbol in symbols]
        init_fcoords = np.array(self.structure.frac_coords)

        with open(filename, "w") as f:
            f.write(self.structure.formula + "\n")
            f.write(" 1.00 \n")

            for i in range(3):
                f.write(" {0} {1} {2} \n".format(*self.structure.lattice.matrix[i, :]))

            f.write(" " + " ".join(symbols) + "\n")
            f.write(" " + " ".join(natoms) + "\n")
            f.write("direct\n")
            for fcoord in init_fcoords:
                f.write(" {0:.8f}  {1:.8f}  {2:.8f} \n".format(*fcoord))

            f.write(" \n")
            f.write(" {0} {1} {2} \n".format(*self.lens))

            for i in range(self.lens[2]):
                for j in range(self.lens[1]):
                    for k in range(self.lens[0]):
                        f.write(" {0:.10e} ".format(self.Pr[k, j, i] * VolinAu))
                        if count % 5 == 0:
                            f.write("\n")
                        count += 1

        f.close()


class SiteOccupancyAnalyzer:
    """
    A class that analyzes the site occupancy of given species using MD trajectories.
    The occupancy of a site is determined based on the shortest distance scheme.

    .. attribute:: site_occ
        N x 1 numpy array that stores the occupancy of all sites associated with
        species. It has the same sequence as the given list of indices.

    .. attribute:: coords_ref
        N x 3 numpy array of fractional coordinates of all reference sites.

    .. attribute:: nsites
        Number of reference sites.

    .. attribute:: structure
        Initial structure.

    """

    def __init__(self, structure, coords_ref, trajectories, species=("Li", "Na")):
        """
        Args:
            structure (pmg_structure): Initial structure.
            coords_ref (N_s x 3 array): Fractional coordinates of N_s given sites
                at which the site occupancy will be computed.
            trajectories (Ntimesteps x Nions x 3 array): Ionic trajectories from MD
                simulation, where Ntimesteps is the number of time steps in MD
                simulation. Note that the coordinates are fractional.
            species(list of str): list of species that are of interest.
        """

        lattice = structure.lattice
        coords_ref = np.array(coords_ref)
        trajectories = np.array(trajectories)
        count = Counter()

        indices = [i for i, site in enumerate(structure) if site.specie.symbol in species]

        for it in range(len(trajectories)):
            dist_matrix = lattice.get_all_distances(coords_ref, trajectories[it, indices, :])
            labels = np.where(dist_matrix == np.min(dist_matrix, axis=0)[None, :])[0]
            count.update(labels)

        site_occ = np.zeros(len(coords_ref), dtype=np.double)

        for i, n in count.most_common(len(coords_ref)):
            site_occ[i] = n / float(len(trajectories))

        self.structure = structure
        self.coords_ref = coords_ref
        self.species = species
        self.indices = indices
        self.nsites = len(coords_ref)
        self.nsteps = len(trajectories)
        self.site_occ = site_occ

    def get_average_site_occupancy(self, indices):
        """
        Get the average site occupancy over a subset of reference sites.
        """
        return np.sum(self.site_occ[indices]) / len(indices)

    @classmethod
    def from_diffusion_analyzer(cls, coords_ref, diffusion_analyzer, species=("Li", "Na")):

        """
        Create a SiteOccupancyAnalyzer object using a diffusion_analyzer object.

        Args:
            coords_ref (nested list of floats): Fractional coordinates of a list
                of reference sites at which the site occupancy will be computed.
            diffusion_analyzer (DiffusionAnalyzer): A
                pymatgen.analysis.diffusion_analyzer.DiffusionAnalyzer object
            species(list of str): list of species that are of interest.
        """

        trajectories = []

        # Initial structure.
        structure = diffusion_analyzer.structure

        # Drifted corrected ionic trajectories
        for s in diffusion_analyzer.get_drift_corrected_structures():
            trajectories.append(s.frac_coords)

        return SiteOccupancyAnalyzer(structure, coords_ref, trajectories, species)



# coding: utf-8

from pymatgen.core import Structure
from pymatgen.analysis.diffusion.neb.io import (
    MVLCINEBEndPointSet,
    MVLCINEBSet,
    get_endpoints_from_index,
    get_endpoint_dist,
)

import unittest
import os

__author__ = "hat003"

test_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)))


def get_path(path_str, dirname="./"):
    cwd = os.path.abspath(os.path.dirname(__file__))
    path = os.path.join(cwd, dirname, path_str)
    return path


class MVLCINEBEndPointSetTest(unittest.TestCase):

    endpoint = Structure.from_file(get_path("POSCAR0", dirname="io_files"))

    def test_incar(self):
        m = MVLCINEBEndPointSet(self.endpoint)
        incar_string = m.incar.get_string(sort_keys=True, pretty=True)
        incar_expect = """ALGO    =  Fast
EDIFF   =  5e-05
EDIFFG  =  -0.02
ENCUT   =  520
IBRION  =  2
ICHARG  =  1
ISIF    =  2
ISMEAR  =  0
ISPIN   =  2
ISYM    =  0
LCHARG  =  False
LORBIT  =  11
LREAL   =  Auto
LWAVE   =  False
MAGMOM  =  35*0.6
NELM    =  200
NELMIN  =  4
NSW     =  99
PREC    =  Accurate
SIGMA   =  0.05"""
        self.assertEqual(incar_string, incar_expect)

    def test_incar_user_setting(self):
        user_incar_settings = {
            "ALGO": "Normal",
            "EDIFFG": -0.05,
            "NELECT": 576,
            "NPAR": 4,
            "NSW": 100,
        }
        m = MVLCINEBEndPointSet(self.endpoint, user_incar_settings=user_incar_settings)
        incar_string = m.incar.get_string(sort_keys=True)
        incar_expect = """ALGO = Normal
EDIFF = 5e-05
EDIFFG = -0.05
ENCUT = 520
IBRION = 2
ICHARG = 1
ISIF = 2
ISMEAR = 0
ISPIN = 2
ISYM = 0
LCHARG = False
LORBIT = 11
LREAL = Auto
LWAVE = False
MAGMOM = 35*0.6
NELECT = 576
NELM = 200
NELMIN = 4
NPAR = 4
NSW = 100
PREC = Accurate
SIGMA = 0.05"""
        self.assertEqual(incar_string.strip(), incar_expect.strip())


class MVLCINEBSetTest(unittest.TestCase):

    structures = [Structure.from_file(get_path("POSCAR" + str(i), dirname="io_files")) for i in range(3)]

    def test_incar(self):
        m = MVLCINEBSet(self.structures)

        incar_string = m.incar.get_string(sort_keys=True)
        incar_expect = """ALGO = Fast
EDIFF = 5e-05
EDIFFG = -0.02
ENCUT = 520
IBRION = 3
ICHAIN = 0
ICHARG = 1
IMAGES = 1
IOPT = 1
ISIF = 2
ISMEAR = 0
ISPIN = 2
ISYM = 0
LCHARG = False
LCLIMB = True
LORBIT = 0
LREAL = Auto
LWAVE = False
MAGMOM = 35*0.6
NELM = 200
NELMIN = 6
NSW = 200
POTIM = 0
PREC = Accurate
SIGMA = 0.05
SPRING = -5"""
        self.assertEqual(incar_string.strip(), incar_expect.strip())

    def test_incar_user_setting(self):
        user_incar_settings = {"IOPT": 3, "EDIFFG": -0.05, "NPAR": 4, "ISIF": 3}
        m = MVLCINEBSet(self.structures, user_incar_settings=user_incar_settings)
        incar_string = m.incar.get_string(sort_keys=True, pretty=True)
        incar_expect = """ALGO    =  Fast
EDIFF   =  5e-05
EDIFFG  =  -0.05
ENCUT   =  520
IBRION  =  3
ICHAIN  =  0
ICHARG  =  1
IMAGES  =  1
IOPT    =  3
ISIF    =  3
ISMEAR  =  0
ISPIN   =  2
ISYM    =  0
LCHARG  =  False
LCLIMB  =  True
LORBIT  =  0
LREAL   =  Auto
LWAVE   =  False
MAGMOM  =  35*0.6
NELM    =  200
NELMIN  =  6
NPAR    =  4
NSW     =  200
POTIM   =  0
PREC    =  Accurate
SIGMA   =  0.05
SPRING  =  -5"""

        self.assertEqual(incar_string.strip(), incar_expect.strip())


class UtilityTest(unittest.TestCase):
    """
    Unit test for outside methods in io.py
    """

    structure = Structure.from_file(get_path("POSCAR", dirname="io_files"))

    def test_get_endpoints_from_index(self):
        endpoints = get_endpoints_from_index(structure=self.structure, site_indices=[0, 1])
        ep_0 = endpoints[0].as_dict()
        ep_1 = endpoints[1].as_dict()
        ep_0_expect = Structure.from_file(get_path("POSCAR_ep0", dirname="io_files")).as_dict()
        ep_1_expect = Structure.from_file(get_path("POSCAR_ep1", dirname="io_files")).as_dict()

        self.assertEqual(ep_0, ep_0_expect)
        self.assertEqual(ep_1, ep_1_expect)

    def test_get_endpoint_dist(self):
        ep0 = Structure.from_file(get_path("POSCAR_ep0", dirname="io_files"))
        ep1 = Structure.from_file(get_path("POSCAR_ep1", dirname="io_files"))
        distances = get_endpoint_dist(ep0, ep1)

        self.assertAlmostEqual(max(distances), 6.3461081051543893, 7)
        self.assertEqual(min(distances), 0.0)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.

from pymatgen.core import Structure, Lattice

__author__ = "Iek-Heng Chu"
__date__ = "01/16"

import unittest
import os
import numpy as np
import matplotlib
from copy import deepcopy

from monty.serialization import loadfn
from pymatgen.analysis.diffusion_analyzer import DiffusionAnalyzer
from pymatgen.analysis.diffusion.aimd.van_hove import (
    VanHoveAnalysis,
    RadialDistributionFunction,
    EvolutionAnalyzer,
)

matplotlib.use("pdf")
tests_dir = os.path.dirname(os.path.abspath(__file__))


class VanHoveTest(unittest.TestCase):
    def test_van_hove(self):
        # Parse the DiffusionAnalyzer object from json file directly
        obj = loadfn(os.path.join(tests_dir, "cNa3PS4_pda.json"))

        vh = VanHoveAnalysis(
            diffusion_analyzer=obj,
            avg_nsteps=5,
            ngrid=101,
            rmax=10.0,
            step_skip=5,
            sigma=0.1,
            species=["Li", "Na"],
        )

        check = np.shape(vh.gsrt) == (20, 101) and np.shape(vh.gdrt) == (20, 101)
        self.assertTrue(check)
        self.assertAlmostEqual(vh.gsrt[0, 0], 3.98942280401, 10)
        self.assertAlmostEqual(vh.gdrt[10, 0], 9.68574868168, 10)


class RDFTest(unittest.TestCase):
    def test_rdf(self):
        # Parse the DiffusionAnalyzer object from json file directly
        obj = loadfn(os.path.join(tests_dir, "cNa3PS4_pda.json"))

        structure_list = []
        for i, s in enumerate(obj.get_drift_corrected_structures()):
            structure_list.append(s)
            if i == 9:
                break
        species = ["Na", "P", "S"]

        # Test from_species
        obj = RadialDistributionFunction.from_species(
            structures=structure_list,
            ngrid=101,
            rmax=10.0,
            cell_range=1,
            sigma=0.1,
            species=species,
            reference_species=species,
        )

        check = np.shape(obj.rdf)[0] == 101 and np.argmax(obj.rdf) == 34
        self.assertTrue(check)
        self.assertAlmostEqual(obj.rdf.max(), 1.634448, 4)

        # Test init
        s = structure_list[0]
        indices = [i for (i, site) in enumerate(s) if site.species_string in species]
        obj = RadialDistributionFunction(
            structures=structure_list,
            ngrid=101,
            rmax=10.0,
            cell_range=1,
            sigma=0.1,
            indices=indices,
            reference_indices=indices,
        )

        check = np.shape(obj.rdf)[0] == 101 and np.argmax(obj.rdf) == 34
        self.assertTrue(check)
        self.assertAlmostEqual(obj.rdf.max(), 1.634448, 4)

        # Test init using structures w/ different lattices
        s0 = deepcopy(structure_list[0])
        sl_1 = [s0, s0, s0 * [1, 2, 1]]
        sl_2 = [s0 * [2, 1, 1], s0, s0]

        obj_1 = RadialDistributionFunction(
            structures=sl_1,
            ngrid=101,
            rmax=10.0,
            cell_range=1,
            sigma=0.1,
            indices=indices,
            reference_indices=indices,
        )
        obj_2 = RadialDistributionFunction(
            structures=sl_2,
            ngrid=101,
            rmax=10.0,
            cell_range=1,
            sigma=0.1,
            indices=indices,
            reference_indices=indices,
        )

        self.assertAlmostEqual(obj_1.rho, obj_2.rho)
        self.assertAlmostEqual(obj_1.rdf[0], obj_2.rdf[0])

    def test_rdf_coordination_number(self):
        # create a simple cubic lattice
        coords = np.array([[0.5, 0.5, 0.5]])
        atom_list = ["S"]
        lattice = Lattice.from_parameters(a=1.0, b=1.0, c=1.0, alpha=90, beta=90, gamma=90)
        structure = Structure(lattice, atom_list, coords)
        rdf = RadialDistributionFunction.from_species(
            structures=[structure], species=["S"], rmax=5.0, sigma=0.1, ngrid=500
        )
        self.assertEqual(rdf.coordination_number[100], 6.0)

    def test_rdf_two_species_coordination_number(self):
        # create a structure with interpenetrating simple cubic lattice
        coords = np.array([[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]])
        atom_list = ["S", "Zn"]
        lattice = Lattice.from_parameters(a=1.0, b=1.0, c=1.0, alpha=90, beta=90, gamma=90)
        structure = Structure(lattice, atom_list, coords)
        rdf = RadialDistributionFunction.from_species(
            structures=[structure],
            species=["S"],
            reference_species=["Zn"],
            rmax=5.0,
            sigma=0.1,
            ngrid=500,
        )
        self.assertEqual(rdf.coordination_number[100], 8.0)

    def setUp(self):
        coords = np.array([[0.5, 0.5, 0.5]])
        atom_list = ["S"]
        lattice = Lattice.from_parameters(a=1.0, b=1.0, c=1.0, alpha=90, beta=90, gamma=90)
        self.structure = Structure(lattice, atom_list, coords)

    def test_raises_valueerror_if_ngrid_is_less_than_2(self):
        with self.assertRaises(ValueError):
            rdf = RadialDistributionFunction.from_species(structures=[self.structure], ngrid=1)

    def test_raises_ValueError_if_sigma_is_not_positive(self):
        with self.assertRaises(ValueError):
            rdf = RadialDistributionFunction.from_species(structures=[self.structure], sigma=0)

    def test_raises_ValueError_if_species_not_in_structure(self):
        with self.assertRaises(ValueError):
            rdf = RadialDistributionFunction.from_species(structures=[self.structure], species=["Cl"])

    def test_raises_ValueError_if_reference_species_not_in_structure(self):
        with self.assertRaises(ValueError):
            rdf = RadialDistributionFunction.from_species(
                structures=[self.structure], species=["S"], reference_species=["Cl"]
            )


class EvolutionAnalyzerTest(unittest.TestCase):
    def test_get_df(self):
        # Parse the DiffusionAnalyzer object from json file directly
        obj = loadfn(os.path.join(tests_dir, "cNa3PS4_pda.json"))

        structure_list = []
        for i, s in enumerate(obj.get_drift_corrected_structures()):
            structure_list.append(s)
            if i == 9:
                break
        eva = EvolutionAnalyzer(structure_list, rmax=10, step=1, time_step=2)
        rdf = eva.get_df(EvolutionAnalyzer.rdf, pair=("Na", "Na"))
        atom_dist = eva.get_df(EvolutionAnalyzer.atom_dist, specie="Na", direction="c")
        check = np.shape(rdf) == (10, 101) and np.shape(atom_dist) == (10, 101) and ("Na", "Na") in eva.pairs
        self.assertTrue(check)
        self.assertAlmostEqual(max(np.array(rdf)[0]), 1.772465, 4)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.

"""
Van Hove analysis for correlations.
"""

import itertools
from collections import Counter
from typing import List, Tuple, Union, Callable

import matplotlib.pyplot as plt
import numpy as np
import pandas as pds
from scipy.stats import norm
from pymatgen.core import Structure
from pymatgen.analysis.diffusion_analyzer import DiffusionAnalyzer
from pymatgen.util.plotting import pretty_plot
from pymatgen.util.typing import ArrayLike


from .rdf import RadialDistributionFunction

__author__ = "Iek-Heng Chu"
__version__ = "1.0"
__date__ = "Aug 9, 2017"


class VanHoveAnalysis:
    """
    Class for van Hove function analysis. In particular, self-part (Gs) and
    distinct-part (Gd) of the van Hove correlation function G(r,t)
    for given species and given structure are computed. If you use this class,
    please consider citing the following paper:

    Zhu, Z.; Chu, I.-H.; Deng, Z. and Ong, S. P. "Role of Na+ Interstitials
    and Dopants in Enhancing the Na+ Conductivity of the Cubic Na3PS4
    Superionic Conductor". Chem. Mater. (2015), 27, pp 8318–8325
    """

    def __init__(
        self,
        diffusion_analyzer: DiffusionAnalyzer,
        avg_nsteps: int = 50,
        ngrid: int = 101,
        rmax: float = 10.0,
        step_skip: int = 50,
        sigma: float = 0.1,
        cell_range: int = 1,
        species: Union[Tuple, List] = ("Li", "Na"),
        reference_species: Union[Tuple, List] = None,
        indices: List = None,
    ):
        """
        Initiation.

        Args:
            diffusion_analyzer (DiffusionAnalyzer): A
                pymatgen.analysis.diffusion_analyzer.DiffusionAnalyzer object
            avg_nsteps (int): Number of t0 used for statistical average
            ngrid (int): Number of radial grid points
            rmax (float): Maximum of radial grid (the minimum is always set zero)
            step_skip (int): # of time steps skipped during analysis. It defines
                the resolution of the reduced time grid
            sigma (float): Smearing of a Gaussian function
            cell_range (int): Range of translational vector elements associated
                with supercell. Default is 1, i.e. including the adjacent image
                cells along all three directions.
            species ([string]): a list of specie symbols of interest.
            reference_species ([string]): Set this option along with 'species'
                parameter to calculate the distinct-part of van Hove function.
                Note that the self-part of van Hove function is always computed
                only for those in "species" parameter.
            indices (list of int): If not None, only a subset of atomic indices
                will be selected for the analysis. If this is given, "species"
                parameter will be ignored.
        """

        # initial check
        if step_skip <= 0:
            raise ValueError("skip_step should be >=1!")

        n_ions, nsteps, ndim = diffusion_analyzer.disp.shape

        if nsteps <= avg_nsteps:
            raise ValueError("Number of timesteps is too small!")

        ntsteps = nsteps - avg_nsteps

        if ngrid - 1 <= 0:
            raise ValueError("Ntot should be greater than 1!")

        if sigma <= 0.0:
            raise ValueError("sigma should be > 0!")

        dr = rmax / (ngrid - 1)
        interval = np.linspace(0.0, rmax, ngrid)
        reduced_nt = int(ntsteps / float(step_skip)) + 1

        lattice = diffusion_analyzer.structure.lattice
        structure = diffusion_analyzer.structure

        if indices is None:
            indices = [j for j, site in enumerate(structure) if site.specie.symbol in species]

        ref_indices = indices
        if reference_species:
            ref_indices = [j for j, site in enumerate(structure) if site.specie.symbol in reference_species]

        rho = float(len(indices)) / lattice.volume

        # reduced time grid
        rtgrid = np.arange(0.0, reduced_nt)
        # van Hove functions
        gsrt = np.zeros((reduced_nt, ngrid), dtype=np.double)
        gdrt = np.zeros((reduced_nt, ngrid), dtype=np.double)

        tracking_ions = []  # type: ArrayLike
        ref_ions = []  # type: ArrayLike

        # auxiliary factor for 4*\pi*r^2
        aux_factor = 4.0 * np.pi * interval ** 2
        aux_factor[0] = np.pi * dr ** 2

        for i, ss in enumerate(diffusion_analyzer.get_drift_corrected_structures()):
            all_fcoords = np.array(ss.frac_coords)
            tracking_ions.append(all_fcoords[indices, :])
            ref_ions.append(all_fcoords[ref_indices, :])

        tracking_ions = np.array(tracking_ions)
        ref_ions = np.array(ref_ions)

        gaussians = norm.pdf(interval[:, None], interval[None, :], sigma) / float(avg_nsteps) / float(len(ref_indices))

        # calculate self part of van Hove function
        image = np.array([0, 0, 0])
        for it in range(reduced_nt):
            dns = Counter()  # type: ignore
            it0 = min(it * step_skip, ntsteps)
            for it1 in range(avg_nsteps):
                dists = [
                    lattice.get_distance_and_image(tracking_ions[it1][u], tracking_ions[it0 + it1][u], jimage=image)[0]
                    for u in range(len(indices))
                ]

                r_indices = [int(dist / dr) for dist in filter(lambda e: e < rmax, dists)]
                dns.update(r_indices)  # type: ignore

            for indx, dn in dns.most_common(ngrid):
                gsrt[it, :] += gaussians[indx, :] * dn

        # calculate distinct part of van Hove function of species
        r = np.arange(-cell_range, cell_range + 1)
        arange = r[:, None] * np.array([1, 0, 0])[None, :]
        brange = r[:, None] * np.array([0, 1, 0])[None, :]
        crange = r[:, None] * np.array([0, 0, 1])[None, :]
        images = arange[:, None, None] + brange[None, :, None] + crange[None, None, :]
        images = images.reshape((len(r) ** 3, 3))

        # find the zero image vector
        zd = np.sum(images ** 2, axis=1)
        indx0 = np.argmin(zd)

        for it in range(reduced_nt):
            dns = Counter()
            it0 = min(it * step_skip, ntsteps)

            for it1 in range(avg_nsteps):
                dcf = (
                    tracking_ions[it0 + it1, :, None, None, :]
                    + images[None, None, :, :]
                    - ref_ions[it1, None, :, None, :]
                )
                dcc = lattice.get_cartesian_coords(dcf)
                d2 = np.sum(dcc ** 2, axis=3)
                dists = [
                    d2[u, v, j] ** 0.5
                    for u in range(len(indices))
                    for v in range(len(ref_indices))
                    for j in range(len(r) ** 3)
                    if u != v or j != indx0
                ]

                r_indices = [int(dist / dr) for dist in filter(lambda e: e < rmax, dists)]
                dns.update(r_indices)

            for indx, dn in dns.most_common(ngrid):
                gdrt[it, :] += gaussians[indx, :] * dn / aux_factor[indx] / rho

        self.obj = diffusion_analyzer
        self.avg_nsteps = avg_nsteps
        self.step_skip = step_skip
        self.rtgrid = rtgrid
        self.interval = interval
        self.gsrt = gsrt
        self.gdrt = gdrt

        # time interval (in ps) in gsrt and gdrt.
        self.timeskip = self.obj.time_step * self.obj.step_skip * step_skip / 1000.0

    def get_3d_plot(self, figsize: Tuple = (12, 8), mode: str = "distinct"):
        """
        Plot 3D self-part or distinct-part of van Hove function, which is
        specified by the input argument 'type'.
        """

        assert mode in ["distinct", "self"]

        if mode == "distinct":
            grt = self.gdrt.copy()
            vmax = 4.0
            cb_ticks = [0, 1, 2, 3, 4]
            cb_label = "$G_d$($t$,$r$)"
        elif mode == "self":
            grt = self.gsrt.copy()
            vmax = 1.0
            cb_ticks = [0, 1]
            cb_label = "4$\pi r^2G_s$($t$,$r$)"

        y = np.arange(np.shape(grt)[1]) * self.interval[-1] / float(len(self.interval) - 1)
        x = np.arange(np.shape(grt)[0]) * self.timeskip
        X, Y = np.meshgrid(x, y, indexing="ij")

        ticksize = int(figsize[0] * 2.5)

        plt.figure(figsize=figsize, facecolor="w")
        plt.xticks(fontsize=ticksize)
        plt.yticks(fontsize=ticksize)

        labelsize = int(figsize[0] * 3)

        plt.pcolor(X, Y, grt, cmap="jet", vmin=grt.min(), vmax=vmax)
        plt.xlabel("Time (ps)", size=labelsize)
        plt.ylabel("$r$ ($\AA$)", size=labelsize)
        plt.axis([x.min(), x.max(), y.min(), y.max()])

        cbar = plt.colorbar(ticks=cb_ticks)
        cbar.set_label(label=cb_label, size=labelsize)
        cbar.ax.tick_params(labelsize=ticksize)
        plt.tight_layout()

        return plt

    def get_1d_plot(self, mode: str = "distinct", times: List = [0.0], colors: List = None):
        """
        Plot the van Hove function at given r or t.

        Args:
            mode (str): Specify which part of van Hove function to be plotted.
            times (list of float): Time moments (in ps) in which the van Hove
                            function will be plotted.
            colors (list strings/tuples): Additional color settings. If not set,
                            seaborn.color_plaette("Set1", 10) will be used.
        """
        if colors is None:
            import seaborn as sns

            colors = sns.color_palette("Set1", 10)

        assert mode in ["distinct", "self"]
        assert len(times) <= len(colors)

        if mode == "distinct":
            grt = self.gdrt.copy()
            ylabel = "$G_d$($t$,$r$)"
            ylim = [-0.005, 4.0]
        elif mode == "self":
            grt = self.gsrt.copy()
            ylabel = "4$\pi r^2G_s$($t$,$r$)"
            ylim = [-0.005, 1.0]

        plt = pretty_plot(12, 8)

        for i, time in enumerate(times):
            index = int(np.round(time / self.timeskip))
            index = min(index, np.shape(grt)[0] - 1)
            new_time = index * self.timeskip
            label = str(new_time) + " ps"
            plt.plot(self.interval, grt[index], color=colors[i], label=label, linewidth=4.0)

        plt.xlabel(r"$r$ ($\AA$)")
        plt.ylabel(ylabel)
        plt.legend(loc="upper right", fontsize=36)
        plt.xlim(0.0, self.interval[-1] - 1.0)
        plt.ylim(ylim[0], ylim[1])
        plt.tight_layout()

        return plt


class EvolutionAnalyzer:
    """
    Analyze the evolution of structures during AIMD simulations.
    """

    def __init__(self, structures: List, rmax: float = 10, step: int = 1, time_step: int = 2):
        """
        Initialization the EvolutionAnalyzer from MD simulations. From the
        structures obtained from MD simulations, we can analyze the structure
        evolution with time by some quantitative characterization such as RDF
        and atomic distribution.

        If you use this class, please consider citing the following paper:
        Tang, H.; Deng, Z.; Lin, Z.; Wang, Z.; Chu, I-H; Chen, C.; Zhu, Z.;
        Zheng, C.; Ong, S. P. "Probing Solid–Solid Interfacial Reactions in
        All-Solid-State Sodium-Ion Batteries with First-Principles
        Calculations", Chem. Mater. (2018), 30(1), pp 163-173.

        Args:
            structures ([Structure]): The list of structures obtained from MD
                simulations.
            rmax (float): Maximum of radial grid (the minimum is always zero).
            step (int): indicate the interval of input structures, which is used
                to calculated correct time step.
            time_step(int): the time step in fs, POTIM in INCAR.
        """
        self.pairs = self.get_pairs(structures[0])
        self.structure = structures[0]
        self.structures = structures
        self.rmax = rmax
        self.step = step
        self.time_step = time_step

    @staticmethod
    def get_pairs(structure: Structure):
        """
        Get all element pairs in a structure.

        Args:
            structure (Structure): structure

        Returns:
            list of tuples
        """
        specie_list = [s.name for s in structure.types_of_specie]
        pairs = itertools.combinations_with_replacement(specie_list, 2)

        return list(pairs)

    @staticmethod
    def rdf(structure: Structure, pair: Tuple, ngrid: int = 101, rmax: float = 10):
        """
        Process rdf from a given structure and pair.

        Args:
            structure (Structure): input structure.
            pair (str tuple): e.g. ("Na", "Na").
            ngrid (int): Number of radial grid points.
            rmax (float): Maximum of radial grid (the minimum is always zero).

        Returns:
            rdf (np.array)
        """
        r = RadialDistributionFunction.from_species(
            [structure],
            ngrid=ngrid,
            rmax=rmax,
            sigma=0.1,
            species=(pair[0]),
            reference_species=(pair[1]),
        )

        return r.rdf

    @staticmethod
    def atom_dist(
        structure: Structure,
        specie: str,
        ngrid: int = 101,
        window: float = 1,
        direction: str = "c",
    ):
        """
        Get atomic distribution for a given specie.

        Args:
            structure (Structure): input structure
            specie (str): species string for an element
            ngrid (int): Number of radial grid points.
            window (float): number of atoms will be counted within the range
                (i-window, i+window), unit is angstrom.
            direction (str): Choose from "a", "b" and "c". Default is "c".

        Returns:
            density (np.array): atomic concentration along one direction.
        """
        if direction in ["a", "b", "c"]:
            l = structure.lattice.__getattribute__(direction)
            ind = ["a", "b", "c"].index(direction)
            assert window <= l, "Window range exceeds valid bounds!"
        else:
            raise ValueError("Choose from a, b and c!")

        atom_list = [site for site in structure.sites if site.species_string == specie]
        atom_total = structure.composition[specie]
        density = []

        for i in np.linspace(0, l - window, ngrid):
            atoms = []
            for j in [-1, 0, 1]:
                temp = [s for s in atom_list if i - window < s.coords[ind] % l + l * j < i + window]
                atoms.extend(temp)

            density.append(len(atoms) / atom_total)

        return np.array(density)

    def get_df(self, func: Callable, save_csv: str = None, **kwargs):
        """
        Get the data frame for a given pair. This step would be very slow if
        there are hundreds or more structures to parse.

        Args:
            func (FunctionType): structure to spectrum function. choose from
                rdf (to get radial distribution function, pair required) or
                get_atomic_distribution (to get atomic distribution, specie
                required). Extra parameters can be parsed using kwargs.
                e.g. To get rdf dataframe:
                    df = EvolutionAnalyzer.get_df(
                        func=EvolutionAnalyzer.rdf, pair=("Na", "Na"))
                e.g. To get atomic distribution:
                    df = EvolutionAnalyzer.get_df(
                        func=EvolutionAnalyzer.atom_dist, specie="Na")
            save_csv (str): save pandas DataFrame to csv.

        Returns:
            pandas.DataFrame object: index is the radial distance in Angstrom,
                and column is the time step in ps.
        """
        prop_table = []
        ngrid = kwargs.get("ngrid", 101)
        if func == self.rdf:
            kwargs["rmax"] = self.rmax

        for structure in self.structures:
            prop_table.append(func(structure, **kwargs))

        index = np.arange(len(self.structures)) * self.time_step * self.step / 1000
        columns = np.linspace(0, self.rmax, ngrid)
        df = pds.DataFrame(prop_table, index=index, columns=columns)

        if save_csv is not None:
            df.to_csv(save_csv)

        return df

    @staticmethod
    def get_min_dist(df: pds.DataFrame, tol: float = 1e-10):
        """
        Get the shortest pair distance from the given DataFrame.

        Args:
            df (DataFrame): index is the radial distance in Angstrom, and
                column is the time step in ps.
            tol (float): any float number less than tol is considered as zero.

        Returns:
            The shortest pair distance throughout the table.
        """
        # TODO: Add unittest
        for i, col in enumerate(df.columns):
            min_dist = df.min(axis="index")[i]
            if min_dist > tol:
                return float(col)
        raise RuntimeError("Getting min dist failed.")

    @staticmethod
    def plot_evolution_from_data(
        df: pds.DataFrame,
        x_label: str = None,
        cb_label: str = None,
        cmap=plt.cm.plasma,  # pylint: disable=E1101
    ):
        """
        Plot the evolution with time for a given DataFrame. It can be RDF,
        atomic distribution or other characterization data we might
        implement in the future.

        Args:

            df (pandas.DataFrame): input DataFrame object, index is the radial
                distance in Angstrom, and column is the time step in ps.
            x_label (str): x label
            cb_label (str): color bar label
            cmap (color map): the color map used in heat map.
                cmocean.cm.thermal is recommended
        Returns:
            matplotlib.axes._subplots.AxesSubplot object
        """
        import seaborn as sns

        sns.set_style("white")

        plt.rcParams["axes.linewidth"] = 1.5
        plt.rcParams["font.family"] = "sans-serif"
        plt.rcParams["xtick.labelsize"] = 20
        plt.rcParams["ytick.labelsize"] = 20
        plt.rcParams["xtick.major.pad"] = 10

        fig, ax = plt.subplots(figsize=(12, 8), facecolor="w")
        ax = sns.heatmap(
            df,
            linewidths=0,
            cmap=cmap,
            annot=False,
            cbar=True,
            xticklabels=10,
            yticklabels=25,
            rasterized=True,
        )
        ax.set_ylim(ax.get_ylim()[::-1])
        ax.collections[0].colorbar.set_label(cb_label, fontsize=30)

        plt.xticks(rotation="horizontal")

        plt.xlabel(x_label, fontsize=30)
        plt.ylabel("Time (ps)", fontsize=30)

        plt.yticks(rotation="horizontal")
        plt.tight_layout()

        return plt

    def plot_rdf_evolution(
        self,
        pair: Tuple,
        cmap=plt.cm.plasma,  # pylint: disable=E1101
        df: pds.DataFrame = None,
    ):
        """
        Plot the RDF evolution with time for a given pair.

        Args:
            pair (str tuple): e.g. ("Na", "Na")
            cmap (color map): the color map used in heat map.
                cmocean.cm.thermal is recommended
            df (DataFrame): external data, index is the radial distance in
                Angstrom, and column is the time step in ps.
        Returns:
            matplotlib.axes._subplots.AxesSubplot object
        """
        if df is None:
            df = self.get_df(func=EvolutionAnalyzer.rdf, pair=pair)
        x_label, cb_label = "$r$ ({}-{}) ($\\rm\AA$)".format(*pair), "$g(r)$"
        p = self.plot_evolution_from_data(df=df, x_label=x_label, cb_label=cb_label, cmap=cmap)

        return p

    def plot_atomic_evolution(
        self,
        specie: str,
        direction: str = "c",
        cmap=plt.cm.Blues,  # pylint: disable=E1101
        df: pds.DataFrame = None,
    ):
        """
        Plot the atomic distribution evolution with time for a given species.

        Args:
            specie (str): species string for an element.
            direction (str): Choose from "a", "b", "c". Default to "c".
            cmap (color map): the color map used in heat map.
            df (DataFrame): external data, index is the atomic distance in
                         Angstrom, and column is the time step in ps.
        Returns:
            matplotlib.axes._subplots.AxesSubplot object
        """
        if df is None:
            df = self.get_df(func=EvolutionAnalyzer.atom_dist, specie=specie, direction=direction)
        x_label, cb_label = (
            "Atomic distribution along {} ".format(direction),
            "Probability",
        )
        p = self.plot_evolution_from_data(df=df, x_label=x_label, cb_label=cb_label, cmap=cmap)
        return p



# coding: utf-8
# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.
"""
Dijkstra's path search on a graph where the nodes are on a periodic graph
"""

__author__ = "Jimmy Shen"
__copyright__ = "Copyright 2021, The Materials Project"
__maintainer__ = "Jimmy Shen"
__email__ = "jmmshn@lbl.gov"
__date__ = "April 11, 2019"

import heapq
import math
from collections import defaultdict
from typing import Dict, Callable, Set

import numpy as np
from networkx.classes.graph import Graph
from pymatgen.analysis.graphs import StructureGraph


def _get_adjacency_with_images(G: Graph) -> Dict:
    """
    Return an adjacency dictionary with properly oriented "to_image" values.
    Note: the current implementation assumes that the original
    "to_jimage" value always corresponds to a an edge u -> v where u <= v.
    Returns:
        dict: Nested dictionary with [start][end][edge_key][data_field]
    """

    def copy_dict(d):
        # recursively copies the dictionary to resolve the fact that
        # two entries in the dictionary can point to the same mutable object
        # eg. changing p_graph[v][u][0]["to_jimage"] also changes
        # p_graph[u][v][0]["to_jimage"] if G was an undirecte graph.
        if isinstance(d, dict):
            new_d = {}
            for k, v in d.items():
                new_d[k] = copy_dict(v)
            return new_d
        return d

    # the dictionary generated by this are inherently linked so we have to
    # recursively copy the data
    p_graph = copy_dict(dict(G.adjacency()))

    # Make sure all the to_jimages are pointing in the correct direction
    for u in p_graph.keys():
        for v in p_graph[u].keys():
            for k, d in p_graph[u][v].items():
                if u > v:
                    p_graph[u][v][k]["to_jimage"] = tuple(np.multiply(-1, d["to_jimage"]))
    return p_graph


def periodic_dijkstra(
    G: Graph,
    sources: set,
    weight: str = "weight",
    max_image: int = 2,
    target_reached: Callable = lambda idx, jimage: False,
):
    """
    Find the lowest cost pathway from a source point in the periodic graph.
    Since the search can move many cells away without finding the target
    we have to limit how many cells away from (0,0,0) to search.

    Args:
        G (Graph): The graph object with additional "to_jimage" fields to
                indicate edges across periodic images.
        sources (set): the index of the source node
        target (int, optional): The index of of target node, if None populate all nodes. Defaults to None.
        max_image (int, optional): Defaults to 3.
        target_reached (callable, optional): A function of (site_index, jimage) used to check
            for stop iteration. This function is always called on the top of heap so it might miss the optimal path but
            often can find a "good enough" path very quickly.
    Returns:
        best_ans: a dictionary of the best cost found to periodic node keyed by (site_index, jimage)
        path_parent: dictionary of optimal path parent for each node given in index-image pairs.
    """

    conn_dict = _get_adjacency_with_images(G.to_undirected())

    # use a default dict since we don't know how far out to search
    best_ans = defaultdict(lambda: math.inf)  # type: ignore

    path_parent = dict()  # the parent of the current node in the optimal path
    pq = []  # type: ignore
    for isource in sources:
        heapq.heappush(pq, (0, (isource, (0, 0, 0))))
    while pq:
        min_val, (cur_idx, cur_image) = heapq.heappop(pq)
        if target_reached(cur_idx, cur_image):
            return best_ans
        if min_val < best_ans[(cur_idx, cur_image)]:
            best_ans[(cur_idx, cur_image)] = min_val
        for next_node, keyed_data in conn_dict[cur_idx].items():
            for k, d in keyed_data.items():
                # get the node index, image pair
                new_image = tuple(np.add(cur_image, d["to_jimage"]))
                next_index_pair = (next_node, new_image)

                if any(abs(i_) > max_image for i_ in new_image):
                    continue

                new_cost = min_val + d[weight]

                if new_cost < best_ans[next_index_pair]:
                    best_ans[next_index_pair] = new_cost
                    path_parent[next_index_pair] = (cur_idx, cur_image)
                    heapq.heappush(pq, (new_cost, next_index_pair))

    return best_ans, path_parent


def periodic_dijkstra_on_sgraph(
    sgraph: StructureGraph,
    sources: Set,
    weight: str = "weight",
    max_image: int = 1,
    target_reached: Callable = lambda idx, jimage: False,
):
    """
    Find the lowest cost pathway from a source point in the periodic graph.
    Since the search can move many cells away without finding the target
    we have to limit how many cells away from (0,0,0) to search.

    Args:
        sgraph (Graph): The StructureGraph object used for path searching
        sources (set): the index of the source node
        target (int, optional): The index of of target node, if None populate all nodes. Defaults to None.
        max_image (int, optional): Defaults to 3.
        target_reached (callable, optional): A function of (site_index, jimage) used to check
            for stop iteration. This function is always called on the top of heap so it might miss the optimal path but
            often can find a "good enough" path very quickly.
    Returns:
        best_ans: a dictionary of the best cost found to periodic node keyed by (site_index, jimage)
        path_parent: dictionary of optimal path parent for each node given in index-image pairs.
    """
    G = sgraph.graph.to_undirected()
    best_ans, path_parent = periodic_dijkstra(
        G,
        sources=sources,
        weight=weight,
        max_image=max_image,
        target_reached=target_reached,
    )
    return best_ans, path_parent


def get_optimal_pathway_rev(path_parent: dict, leaf_node: tuple):
    """
    follow a leaf node all the way up to source.
    """
    cur = leaf_node
    while cur in path_parent:
        yield cur
        cur = path_parent[cur]
    yield cur



import os
import unittest

from monty.serialization import loadfn
import numpy as np
from pymatgen.core import Lattice, Structure

from pymatgen.analysis.diffusion.aimd.rdf import RadialDistributionFunctionFast


tests_dir = os.path.dirname(os.path.abspath(__file__))


class RDFTest(unittest.TestCase):
    def test_rdf(self):
        # Parse the DiffusionAnalyzer object from json file directly
        obj = loadfn(os.path.join(tests_dir, "cNa3PS4_pda.json"))

        structure_list = []
        for i, s in enumerate(obj.get_drift_corrected_structures()):
            structure_list.append(s)
            if i == 9:
                break
        species = ["Na", "P", "S"]

        # Test from_species
        obj = RadialDistributionFunctionFast(structures=structure_list, ngrid=101, rmax=10.0, sigma=0.1)

        r, s_na_rdf = obj.get_rdf("S", "Na")
        self.assertTrue(s_na_rdf.shape == (101,))

        self.assertAlmostEqual(r[np.argmax(s_na_rdf)], 2.9000, 4)

    def test_rdf_coordination_number(self):
        # create a simple cubic lattice
        coords = np.array([[0.5, 0.5, 0.5]])
        atom_list = ["S"]
        lattice = Lattice.from_parameters(a=1.0, b=1.0, c=1.0, alpha=90, beta=90, gamma=90)
        structure = Structure(lattice, atom_list, coords)
        rdf = RadialDistributionFunctionFast(structures=[structure], rmax=5.0, sigma=0.01, ngrid=500)
        self.assertEqual(np.round(rdf.get_coordination_number("S", "S")[1][110], 2), 6.0)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.
import os
import unittest

import numpy as np
from monty.serialization import loadfn
from pymatgen.core import Structure, PeriodicSite
from pymatgen.io.vasp import Chgcar

from pymatgen.analysis.diffusion.neb.full_path_mapper import (
    ChargeBarrierGraph,
    MigrationGraph,
    MigrationHop,
    get_hop_site_sequence,
)
from pymatgen.analysis.diffusion.neb.periodic_dijkstra import _get_adjacency_with_images

dir_path = os.path.dirname(os.path.realpath(__file__))

__author__ = "Jimmy Shen"
__version__ = "1.0"
__date__ = "April 10, 2019"


class MigrationGraphSimpleTest(unittest.TestCase):
    def setUp(self):
        struct = Structure.from_file(f"{dir_path}/full_path_files/MnO2_full_Li.vasp")
        self.fpm = MigrationGraph.with_distance(structure=struct, migrating_specie="Li", max_distance=4)

    def test_get_pos_and_migration_path(self):
        """
        Make sure that we can populate the graph with MigrationHop Objects
        """
        self.fpm._get_pos_and_migration_path(0, 1, 1)
        self.assertAlmostEqual(self.fpm.migration_graph.graph[0][1][1]["hop"].length, 3.571248, 4)

    def test_get_summary_dict(self):
        summary_dict = self.fpm.get_summary_dict()
        self.assertTrue("hop_label", summary_dict["hops"][0])
        self.assertTrue("hop_label", summary_dict["unique_hops"][0])


class MigrationGraphFromEntriesTest(unittest.TestCase):
    def setUp(self):
        self.test_ents_MOF = loadfn(f"{dir_path}/full_path_files/Mn6O5F7_cat_migration.json")
        self.aeccar_MOF = Chgcar.from_file(f"{dir_path}/full_path_files/AECCAR_Mn6O5F7.vasp")
        self.li_ent = loadfn(f"{dir_path}/full_path_files/li_ent.json")["li_ent"]

        self.full_struct = MigrationGraph.get_structure_from_entries(
            base_entries=[self.test_ents_MOF["ent_base"]],
            inserted_entries=self.test_ents_MOF["one_cation"],
            migrating_ion_entry=self.li_ent,
        )[0]

    def test_migration_graph_construction(self):
        self.assertEqual(self.full_struct.composition["Li"], 8)
        mg = MigrationGraph.with_distance(self.full_struct, migrating_specie="Li", max_distance=4.0)
        self.assertEqual(len(mg.migration_graph.structure), 8)


class MigrationGraphComplexTest(unittest.TestCase):
    def setUp(self):
        struct = Structure.from_file(f"{dir_path}/full_path_files/MnO2_full_Li.vasp")
        self.fpm_li = MigrationGraph.with_distance(structure=struct, migrating_specie="Li", max_distance=4)

        # Particularity difficult pathfinding since both the starting and ending positions are outside the unit cell
        struct = Structure.from_file(f"{dir_path}/full_path_files/Mg_2atom.vasp")
        self.fpm_mg = MigrationGraph.with_distance(structure=struct, migrating_specie="Mg", max_distance=2)

    def test_group_and_label_hops(self):
        """
        Check that the set of end points in a group of similiarly labeled hops are all the same
        """
        edge_labs = np.array([d["hop_label"] for u, v, d in self.fpm_li.migration_graph.graph.edges(data=True)])

        site_labs = np.array(
            [
                (
                    d["hop"].symm_structure.wyckoff_symbols[d["hop"].iindex],
                    d["hop"].symm_structure.wyckoff_symbols[d["hop"].eindex],
                )
                for u, v, d in self.fpm_li.migration_graph.graph.edges(data=True)
            ]
        )

        for itr in range(edge_labs.max()):
            sub_set = site_labs[edge_labs == itr]
            for end_point_labels in sub_set:
                self.assertTrue(sorted(end_point_labels) == sorted(sub_set[0]))

    def test_unique_hops_dict(self):
        """
        Check that the unique hops are inequilvalent
        """
        unique_list = [v for k, v in self.fpm_li.unique_hops.items()]
        all_pairs = [(mg1, mg2) for i1, mg1 in enumerate(unique_list) for mg2 in unique_list[i1 + 1 :]]

        for migration_path in all_pairs:
            self.assertNotEqual(migration_path[0]["hop"], migration_path[1]["hop"])

    def test_add_data_to_similar_edges(self):
        # passing normal data
        self.fpm_li.add_data_to_similar_edges(0, {"key0": "data"})
        for u, v, d in self.fpm_li.migration_graph.graph.edges(data=True):
            if d["hop_label"] == 0:
                self.assertEqual(d["key0"], "data")

        # passing ordered list data
        migration_path = self.fpm_li.unique_hops[1]["hop"]
        self.fpm_li.add_data_to_similar_edges(1, {"key1": [1, 2, 3]}, m_path=migration_path)
        for u, v, d in self.fpm_li.migration_graph.graph.edges(data=True):
            if d["hop_label"] == 1:
                self.assertEqual(d["key1"], [1, 2, 3])

        # passing ordered list with direction
        migration_path_reversed = MigrationHop(
            isite=migration_path.esite,
            esite=migration_path.isite,
            symm_structure=migration_path.symm_structure,
        )
        self.fpm_li.add_data_to_similar_edges(2, {"key2": [1, 2, 3]}, m_path=migration_path_reversed)
        for u, v, d in self.fpm_li.migration_graph.graph.edges(data=True):
            if d["hop_label"] == 2:
                self.assertEqual(d["key2"], [3, 2, 1])

    def test_assign_cost_to_graph(self):
        self.fpm_li.assign_cost_to_graph()  # use 'hop_distance'
        for u, v, d in self.fpm_li.migration_graph.graph.edges(data=True):
            self.assertAlmostEqual(d["cost"], d["hop_distance"], 4)

        self.fpm_li.assign_cost_to_graph(cost_keys=["hop_distance", "hop_distance"])
        for u, v, d in self.fpm_li.migration_graph.graph.edges(data=True):
            self.assertAlmostEqual(d["cost"], d["hop_distance"] ** 2, 4)

    def test_periodic_dijkstra(self):
        self.fpm_li.assign_cost_to_graph()  # use 'hop_distance'

        # test the connection graph
        sgraph = self.fpm_li.migration_graph
        G = sgraph.graph.to_undirected()
        conn_dict = _get_adjacency_with_images(G)
        for u in conn_dict.keys():
            for v in conn_dict[u]:
                for k, d in conn_dict[u][v].items():
                    neg_image = tuple(-dim_ for dim_ in d["to_jimage"])
                    opposite_connections = [d2_["to_jimage"] for k2_, d2_ in conn_dict[v][u].items()]
                    self.assertIn(neg_image, opposite_connections)

    def test_get_path(self):
        self.fpm_li.assign_cost_to_graph()  # use 'hop_distance'
        paths = [*self.fpm_li.get_path()]
        p_strings = {"->".join(map(str, get_hop_site_sequence(ipath, start_u=u))) for u, ipath in paths}
        self.assertIn("5->7->5", p_strings)
        # convert each pathway to a string representation
        paths = [*self.fpm_li.get_path(max_val=2.0)]
        p_strings = {"->".join(map(str, get_hop_site_sequence(ipath, start_u=u))) for u, ipath in paths}

        # After checking trimming the graph more hops are needed for the same path
        self.assertIn("5->3->7->2->5", p_strings)

        self.fpm_mg.assign_cost_to_graph()  # use 'hop_distance'
        paths = [*self.fpm_mg.get_path()]
        p_strings = {"->".join(map(str, get_hop_site_sequence(ipath, start_u=u))) for u, ipath in paths}
        self.assertIn("1->0->1", p_strings)

    def test_not_matching_first(self):
        structure = Structure.from_file(f"{dir_path}/pathfinder_files/Li6MnO4.json")
        fpm_lmo = MigrationGraph.with_distance(structure, "Li", max_distance=4)
        for u, v, d in fpm_lmo.migration_graph.graph.edges(data=True):
            self.assertIn(d["hop"].eindex, {0, 1})


class ChargeBarrierGraphTest(unittest.TestCase):
    def setUp(self):
        self.full_sites_MOF = loadfn(f"{dir_path}/full_path_files/LixMn6O5F7_full_sites.json")
        self.aeccar_MOF = Chgcar.from_file(f"{dir_path}/full_path_files/AECCAR_Mn6O5F7.vasp")
        self.cbg = ChargeBarrierGraph.with_distance(
            structure=self.full_sites_MOF,
            migrating_specie="Li",
            max_distance=4,
            potential_field=self.aeccar_MOF,
            potential_data_key="total",
        )
        self.cbg._tube_radius = 10000

    def test_integration(self):
        """
        Sanity check: for a long enough diagonaly hop, if we turn the radius of the tube way up, it should cover the entire unit cell
        """
        total_chg_per_vol = (
            self.cbg.potential_field.data["total"].sum()
            / self.cbg.potential_field.ngridpts
            / self.cbg.potential_field.structure.volume
        )
        self.assertAlmostEqual(
            self.cbg._get_chg_between_sites_tube(self.cbg.unique_hops[2]["hop"]),
            total_chg_per_vol,
        )

        self.cbg._tube_radius = 2
        # self.cbg.populate_edges_with_chg_density_info()

        # find this particular hop
        ipos = [0.33079153, 0.18064031, 0.67945924]
        epos = [0.33587514, -0.3461259, 1.15269302]
        isite = PeriodicSite("Li", ipos, self.cbg.structure.lattice)
        esite = PeriodicSite("Li", epos, self.cbg.structure.lattice)
        ref_hop = MigrationHop(isite=isite, esite=esite, symm_structure=self.cbg.symm_structure)
        hop_idx = -1
        for k, d in self.cbg.unique_hops.items():
            if d["hop"] == ref_hop:
                hop_idx = k

        self.assertAlmostEqual(
            self.cbg._get_chg_between_sites_tube(self.cbg.unique_hops[hop_idx]["hop"]),
            0.188952739835188,
            3,
        )

    def test_populate_edges_with_chg_density_info(self):
        """
        Test that all of the sites with similar lengths have similar charge densities,
        this will not always be true, but it valid in this Mn6O5F7
        """
        self.cbg.populate_edges_with_chg_density_info()
        length_vs_chg = list(
            sorted([(d["hop"].length, d["chg_total"]) for u, v, d in self.cbg.migration_graph.graph.edges(data=True)])
        )
        prv = None
        for length, chg in length_vs_chg:
            if prv is None:
                prv = (length, chg)
                continue

            if 1.05 > length / prv[0] > 0.95:
                self.assertAlmostEqual(chg, prv[1], 3)

    def test_get_summary_dict(self):
        summary_dict = self.cbg.get_summary_dict()
        self.assertTrue("chg_total", summary_dict["hops"][0])
        self.assertTrue("chg_total", summary_dict["unique_hops"][0])


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.
"""
Functions for querying Materials Project style MongoStores that contains
cathode materials The functions are isolated from the rest of the package so
that the rest of the package will not depend on Maggma
"""

import json
import logging
import zlib
from itertools import chain
from typing import Union

import gridfs
from maggma.stores import MongoStore
from monty.serialization import MontyDecoder
from pymatgen.core import Structure
from pymatgen.entries.compatibility import MaterialsProjectCompatibility
from pymatgen.entries.computed_entries import ComputedStructureEntry

__author__ = "Jimmy Shen"
__copyright__ = "Copyright 2019, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Jimmy Shen"
__email__ = "jmmshn@lbl.gov"
__date__ = "July 21, 2019"

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
compat = MaterialsProjectCompatibility("Advanced")


def get_ent_from_db(
    elec_store: MongoStore,
    material_store: MongoStore,
    tasks_store: MongoStore,
    batt_id: Union[str, int] = None,
    task_id: Union[str, int] = None,
    get_aeccar: bool = False,
    working_ion: str = "Li",
    add_fields: list = None,
    get_initial: bool = False,
):
    """
    Get the migration path information in the form of a ComputedEntryGraph
    object from the an atomate data stack

    The algorithm gets all tasks with structures that are valid (i.e. matches a
    base structure) and generates a migration pathway object using all possible
    relaxed working ion positions found in that set. Since each material entry
    might contain multiple calculations with different cell sizes, this will
    have to work at the task level. Need to group tasks together based on the
    cell size of the base material

    Note that SPGlib is some times inconsistent when it comes to the getting
    the number of symmetry operations for a given structure. Sometimes
    structures that are the same using StructureMatcher.fit will have
    different number of symmetry operation. As such we will check the number
    of operations for each base structure in a given family of structures
    and take the case with the highest number symmetry operations In cases
    where AECCAR is required, only the tasks with AECCARs will have this data.

    Args:

        elec_store: Electrode documents one per each similar group of
            insertion materials, can also use any db that contains a
        material_ids list with topotactic structures
        material_store: Material documenets one per each similar structure (
            multiple tasks)
        tasks_store: Task documents one per each VASP calculation
        batt_id: battery id to lookup in a database.
        task_id: if battery id is not provided then look up a materials id.
        get_aeccar: If True, only find base tasks with the charge density stored
        working_ion: Name of the working ion. Defaults to 'Li'.
        add_fields: Take these fields from the task_documents and store them
            in ComputedStructureEntry
        get_initial: Store the initial structure of a calculation

    """

    task_ids_type = type(material_store.query_one({})["task_ids"][0])
    material_ids_type = type(elec_store.query_one({})["material_ids"][0])

    logger.debug(material_ids_type)

    def get_task_ids_from_batt_id(b_id):
        mat_ids = list(map(task_ids_type, elec_store.query_one({"battid": b_id})["material_ids"]))
        logger.debug(f"mat_ids : {mat_ids}")
        l_task_ids = [imat["task_ids"] for imat in material_store.query({"task_ids": {"$in": mat_ids}})]
        l_task_ids = list(chain.from_iterable(l_task_ids))
        logger.debug(f"l_task_ids : {l_task_ids}")
        return l_task_ids

    def get_batt_ids_from_task_id(t_id):
        l_task_ids = [c0["task_ids"] for c0 in material_store.query({"task_ids": {"$in": [int(t_id)]}})]
        l_task_ids = list(chain.from_iterable(l_task_ids))
        l_task_ids = list(map(material_ids_type, l_task_ids))
        logger.debug(f"l_task_ids : {l_task_ids}")
        l_mat_ids = [c0["material_ids"] for c0 in elec_store.query({"material_ids": {"$in": l_task_ids}})]
        l_mat_ids = list(chain.from_iterable(l_mat_ids))
        l_mat_ids = list(map(task_ids_type, l_mat_ids))
        logger.debug(f"l_mat_ids : {l_mat_ids}")
        l_task_ids = [c0["task_ids"] for c0 in material_store.query({"task_ids": {"$in": l_mat_ids}})]
        l_task_ids = list(chain.from_iterable(l_task_ids))
        logger.debug(f"l_task_ids : {l_task_ids}")
        return l_task_ids

    def get_entry(task_doc, base_with_aeccar=False, add_fields=None, get_initial=None):
        # we don't really need to think about compatibility for now if just
        # want to make a code that automate NEB calculations
        tmp_struct = Structure.from_dict(task_doc["output"]["structure"])
        settings_dict = dict(
            potcar_spec=task_doc["calcs_reversed"][0]["input"]["potcar_spec"],
            rung_type=task_doc["calcs_reversed"][0]["run_type"],
        )
        if "is_hubbard" in task_doc["calcs_reversed"][0].keys():
            settings_dict["hubbards"] = task_doc["calcs_reversed"][0]["hubbards"]
            settings_dict["is_hubbard"] = (task_doc["calcs_reversed"][0]["is_hubbard"],)

        entry = ComputedStructureEntry(
            structure=tmp_struct,
            energy=task_doc["output"]["energy"],
            parameters=settings_dict,
            entry_id=task_doc["task_id"],
        )
        if base_with_aeccar:
            logger.debug("test")
            aec_id = tasks_store.query_one({"task_id": entry.entry_id})["task_id"]
            aeccar = get_aeccar_from_store(tasks_store, aec_id)
            entry.data.update({"aeccar": aeccar})

        if add_fields:
            for field in add_fields:
                if field in task_doc:
                    entry.data.update({field: task_doc[field]})
        if get_initial:
            entry.data.update({"initial_structure": task_doc["input"]["structure"]})

        return entry

    # Require a single base entry and multiple inserted entries to populate
    # the migration pathways

    # getting a full list of task ids
    # Batt_id -> material_id -> all task_ids
    # task_id -> mat_ids -> Batt_ids -> material_id -> all task_ids
    if batt_id:
        all_tasks = get_task_ids_from_batt_id(batt_id)
    else:
        all_tasks = get_batt_ids_from_task_id(task_id)
    # get_all the structures
    if get_aeccar:
        all_ents_base = [
            get_entry(
                c0,
                base_with_aeccar=True,
                add_fields=add_fields,
                get_initial=get_initial,
            )
            for c0 in tasks_store.query(
                {
                    "task_id": {"$in": all_tasks},
                    "elements": {"$nin": [working_ion]},
                    "calcs_reversed.0.aeccar0_fs_id": {"$exists": 1},
                }
            )
        ]
    else:
        all_ents_base = [
            get_entry(c0)
            for c0 in tasks_store.query({"task_id": {"$in": all_tasks}, "elements": {"$nin": [working_ion]}})
        ]
    logger.debug(f"Number of base entries: {len(all_ents_base)}")

    all_ents_insert = [
        get_entry(c0, add_fields=add_fields, get_initial=get_initial)
        for c0 in tasks_store.query({"task_id": {"$in": all_tasks}, "elements": {"$in": [working_ion]}})
    ]
    logger.debug(f"Number of inserted entries: {len(all_ents_insert)}")
    tmp = [f"{itr.name}({itr.entry_id})" for itr in all_ents_insert]
    logger.debug(f"{tmp}")
    return all_ents_base, all_ents_insert


def get_aeccar_from_store(tstore, task_id):
    """
    Read the AECCAR grid_fs data into a Chgcar object

    Args:
        tstore (MongoStore): MongoStore for the tasks database
        task_id: The task_id of the material entry

    Returns:
        pymatgen Chrgcar object: The AECCAR data from a given task
    """

    m_task = tstore.query_one({"task_id": task_id})
    try:
        fs_id = m_task["calcs_reversed"][0]["aeccar0_fs_id"]
    except BaseException:
        logger.info("AECCAR0 Missing from task # {}".format(task_id))
        return None

    fs = gridfs.GridFS(tstore._collection.database, "aeccar0_fs")
    chgcar_json = zlib.decompress(fs.get(fs_id).read())
    aeccar0 = json.loads(chgcar_json, cls=MontyDecoder)

    try:
        fs_id = m_task["calcs_reversed"][0]["aeccar2_fs_id"]
    except BaseException:
        logger.info("AECCAR2 Missing from task # {}".format(task_id))
        return None

    fs = gridfs.GridFS(tstore._collection.database, "aeccar2_fs")
    chgcar_json = zlib.decompress(fs.get(fs_id).read())
    aeccar2 = json.loads(chgcar_json, cls=MontyDecoder)
    return aeccar0 + aeccar2



# coding: utf-8
# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.

"""
This module implements clustering algorithms to determine centroids, with
adaption for periodic boundary conditions. This can be used, for example, to
determine likely atomic positions from MD trajectories.
"""

import random
import warnings

import numpy as np
from pymatgen.util.coord import all_distances, pbc_diff

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2013, The Materials Virtual Lab"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "ongsp@ucsd.edu"
__date__ = "3/18/15"


class Kmeans:
    """
    Simple kmeans clustering.
    """

    def __init__(self, max_iterations: int = 1000):
        """
        Args:
            max_iterations (int): Maximum number of iterations to run KMeans algo.
        """
        self.max_iterations = max_iterations

    def cluster(self, points, k, initial_centroids=None):
        """
        Args:
            points (ndarray): Data points as a mxn ndarray, where m is the
                number of features and n is the number of data points.
            k (int): Number of means.
            initial_centroids (np.array): Initial guess for the centroids. If
                None, a randomized array of points is used.

        Returns:
            centroids, labels, ss: centroids are the final centroids, labels
            provide the index for each point, and ss in the final sum squared
            distances.
        """
        centroids = np.array(random.sample(list(points), k)) if initial_centroids is None else initial_centroids

        # Initialize book keeping vars.
        iterations = 0
        old_centroids = None

        # Run the main k-means algorithm
        while not self.should_stop(old_centroids, centroids, iterations):
            # Save old centroids for convergence test. Book keeping.
            old_centroids = centroids
            iterations += 1
            # Assign labels to each datapoint based on centroids
            labels, ss = self.get_labels(points, centroids)

            # Assign centroids based on datapoint labels
            centroids = self.get_centroids(points, labels, k, centroids)

        labels, ss = self.get_labels(points, centroids)
        # We can get the labels too by calling getLabels(dataSet, centroids)
        return centroids, labels, ss

    @staticmethod
    def get_labels(points, centroids):
        """
        For each element in the dataset, chose the closest centroid.
        Make that centroid the element's label.

        Args:
            points: List of points
            centroids: List of centroids
        """
        dists = all_distances(points, centroids)
        min_dists = np.min(dists, axis=1)
        return np.where(dists == min_dists[:, None])[1], np.sum(min_dists ** 2)

    @staticmethod
    def get_centroids(points, labels, k, centroids):
        """
        Each centroid is the geometric mean of the points that
        have that centroid's label. Important: If a centroid is empty (no
        points have that centroid's label) you should randomly re-initialize it.

        Args:
            points: List of points
            labels: List of labels
            k: Number of means
            centroids: List of centroids
        """
        labels = np.array(labels)
        centroids = []
        for i in range(k):
            ind = np.where(labels == i)[0]
            if len(ind) > 0:
                centroids.append(np.average(points[ind, :], axis=0))
            else:
                centroids.append(get_random_centroid(points))
        return np.array(centroids)

    def should_stop(self, old_centroids, centroids, iterations):
        """
        Check for stopping conditions.

        Args:
            old_centroids: List of old centroids
            centroids: List of centroids
            iterations: Number of iterations thus far.
        """
        if iterations > self.max_iterations:
            warnings.warn("Max iterations %d reached!" % self.max_iterations)
            return True
        if old_centroids is None:
            return False
        return np.allclose(old_centroids, centroids)


class KmeansPBC(Kmeans):
    """
    A version of KMeans that work with PBC. Distance metrics have to change,
    as well as new centroid determination. The points supplied should be
    fractional coordinates.
    """

    def __init__(self, lattice, max_iterations=1000):
        """
        Args:
            lattice: Lattice
            max_iterations: Maximum number of iterations to run KMeans.
        """
        self.lattice = lattice
        self.max_iterations = max_iterations

    def get_labels(self, points, centroids):
        """
        For each element in the dataset, chose the closest centroid.
        Make that centroid the element's label.

        Args:
            points: List of points
            centroids: List of centroids
        """
        dists = self.lattice.get_all_distances(points, centroids)
        min_dists = np.min(dists, axis=1)
        return np.where(dists == min_dists[:, None])[1], np.sum(min_dists ** 2)

    def get_centroids(self, points, labels, k, centroids):
        """
        Each centroid is the geometric mean of the points that
        have that centroid's label. Important: If a centroid is empty (no
        points have that centroid's label) you should randomly re-initialize it.

        Args:
            points: List of points
            labels: List of labels
            k: Number of means
            centroids: List of centroids
        """
        m, n = points.shape
        labels = np.array(labels)
        new_centroids = []
        for i in range(k):
            ind = np.where(labels == i)[0]
            if len(ind) > 0:
                c = np.zeros(n)
                for j in ind:
                    dist, image = self.lattice.get_distance_and_image(centroids[i], points[j])
                    c += points[j] + image
                c /= len(ind)
                c = np.mod(c, 1)
            else:
                c = get_random_centroid(points)
                c = np.mod(c, 1)
            new_centroids.append(c)
        return np.array(new_centroids)

    def should_stop(self, old_centroids, centroids, iterations):
        """
        Check for stopping conditions.

        Args:
            old_centroids: List of old centroids
            centroids: List of centroids
            iterations: Number of iterations thus far.
        """
        if iterations > self.max_iterations:
            warnings.warn("Max iterations %d reached!" % self.max_iterations)
            return True
        if old_centroids is None:
            return False
        for c1, c2 in zip(old_centroids, centroids):
            if not np.allclose(pbc_diff(c1, c2), [0, 0, 0]):
                return False
        return True


def get_random_centroid(points):
    """
    Generate a random centroid based on points.

    Args:
        points: List of points.
    """
    m, n = points.shape
    maxd = np.max(points, axis=0)
    mind = np.min(points, axis=0)
    return np.array([random.uniform(mind[i], maxd[i]) for i in range(n)])


def get_random_centroids(points, k):
    """
    Generate k random centroids based on points.

    Args:
        points: List of points.
        k: Number of means.
    """
    centroids = []
    for i in range(k):
        centroids.append(get_random_centroid(points))
    return np.array(centroids)



# coding: utf-8
# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.
"""
Migraiton Graph Analysis
"""

__author__ = "Jimmy Shen"
__copyright__ = "Copyright 2019, The Materials Project"
__maintainer__ = "Jimmy Shen"
__email__ = "jmmshn@lbl.gov"
__date__ = "April 11, 2019"

import logging
import operator
from copy import deepcopy
from itertools import starmap
from typing import Callable, Dict, List, Union

import networkx as nx
import numpy as np
from monty.json import MSONable
from pymatgen.core import Composition
from pymatgen.analysis.graphs import StructureGraph
from pymatgen.analysis.local_env import MinimumDistanceNN, NearNeighbors
from pymatgen.analysis.path_finder import ChgcarPotential, NEBPathfinder
from pymatgen.core import PeriodicSite, Structure
from pymatgen.entries.computed_entries import ComputedEntry, ComputedStructureEntry
from pymatgen.io.vasp import VolumetricData
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.symmetry.structure import SymmetrizedStructure

from pymatgen.analysis.diffusion.neb.pathfinder import MigrationHop
from pymatgen.analysis.diffusion.neb.periodic_dijkstra import (
    get_optimal_pathway_rev,
    periodic_dijkstra,
)
from pymatgen.analysis.diffusion.utils.parse_entries import process_entries

logger = logging.getLogger(__name__)


def generic_groupby(list_in: list, comp: Callable = operator.eq):
    """
    Group a list of unsortable objects
    Args:
        list_in: A list of generic objects
        comp: (Default value = operator.eq) The comparator
    Returns:
        [int] list of labels for the input list
    """
    list_out = [None] * len(list_in)
    label_num = 0
    for i1, ls1 in enumerate(list_out):
        if ls1 is not None:
            continue
        list_out[i1] = label_num  # type: ignore
        for i2, ls2 in list(enumerate(list_out))[i1 + 1 :]:
            if comp(list_in[i1], list_in[i2]):
                if list_out[i2] is None:
                    list_out[i2] = list_out[i1]
                else:
                    list_out[i1] = list_out[i2]
                    label_num -= 1
        label_num += 1
    return list_out


class MigrationGraph(MSONable):
    """
    A python object for handling the migratrion graph of a given base
    structure and mobile specie sites within that base structure.

    Each mobile site is a node in the migration graph and hops
    (between sites) are edges. The migration graph object uses
    the Structure Graph from pymatgen.
    """

    def __init__(
        self,
        structure: Structure,
        migration_graph: StructureGraph,
        symprec=0.1,
        vac_mode=False,
    ):
        """
        Construct the MigrationGraph object using a potential_field will
        all mobile sites occupied. A potential_field graph is generated by
        connecting all sites as specified by the migration graph.
        The sites are decorated with Migration graph objects and then grouped
        together based on their equivalence.
        Args:
            structure: Structure with base framework and mobile sites.
             When used with structure_is_base = True, only the base framework
             structure, does not contain any migrating sites.
            migration_graph: The StructureGraph object that defines the
             migration network
            symprec (float): Symmetry precision to determine equivalence
             of migration events
            vac_mode (Bool): indicates whether vacancy mode should be used
        """
        self.structure = structure
        self.migration_graph = migration_graph
        self.symprec = symprec
        self.vac_mode = vac_mode
        if self.vac_mode:
            raise NotImplementedError("Vacancy mode is not yet implemented")
        # Generate the graph edges between these all the sites
        self.migration_graph.set_node_attributes()  # popagate the sites properties to the graph nodes
        # For poperies like unique_hops we might be interested in modifying them after creation
        # So let's not convert them into properties for now.  (Awaiting rewrite once the usage becomes more clear.)
        self._populate_edges_with_migration_paths()
        self._group_and_label_hops()

    @property
    def only_sites(self) -> Structure:
        """
        A structure that only contains the migrating species
        """
        return self.migration_graph.structure

    @property
    def host_structure(self) -> Structure:
        """
        A structure that only contains the non-migrating species
        """
        host_struct = self.structure.copy()
        rm_sites = set()
        for isite in self.only_sites:
            neighbors_ = host_struct.get_neighbors_in_shell(isite.coords, r=0.0, dr=0.05, include_index=True)
            if len(neighbors_) == 0:
                continue
            for n_ in neighbors_:
                rm_sites.add(n_.index)
        host_struct.remove_sites(list(rm_sites))
        return host_struct

    @property
    def symm_structure(self) -> SymmetrizedStructure:
        """
        The symmetrized structure with the present item's symprec value
        """
        a = SpacegroupAnalyzer(self.structure, symprec=self.symprec)
        sym_struct = a.get_symmetrized_structure()
        if not isinstance(sym_struct, SymmetrizedStructure):
            raise RuntimeError("Symmetrized structure could not be generated.")
        return sym_struct

    @property
    def unique_hops(self):
        """
        The unique hops dictionary keyed by the hop label
        """
        # reversed so that the first instance represents the group of distinct hops
        ihop_data = list(reversed(list(self.migration_graph.graph.edges(data=True))))
        for u, v, d in ihop_data:
            d["iindex"] = u
            d["eindex"] = v
            d["hop_distance"] = d["hop"].length
        return {d["hop_label"]: d for u, v, d in ihop_data}

    @classmethod
    def with_base_structure(
        cls, base_structure: Structure, migration_graph: StructureGraph, **kwargs
    ) -> "MigrationGraph":
        """
        Args:
            base_structure: base framework structure that does not contain any
             migrating sites.
        Returns:
            A constructed MigrationGraph object
        """
        sites = migration_graph.structure.sites + base_structure.sites
        structure = Structure.from_sites(sites)
        return cls(structure=structure, migration_graph=migration_graph, **kwargs)

    @classmethod
    def with_local_env_strategy(
        cls, structure: Structure, migrating_specie: str, nn: NearNeighbors, **kwargs
    ) -> "MigrationGraph":
        """
        Using a specific nn strategy to get the connectivity graph between all the migrating ion sites.
        Args:
            structure: Input structure that contains all sites.
            migrating_specie: The specie that migrates. E.g. "Li".
            nn: The specific local environment object used to connect the migrating ion sites.
        Returns:
            A constructed MigrationGraph object
        """
        only_sites = get_only_sites_from_structure(structure, migrating_specie)
        migration_graph = StructureGraph.with_local_env_strategy(only_sites, nn)
        return cls(structure=structure, migration_graph=migration_graph, **kwargs)

    @classmethod
    def with_distance(
        cls, structure: Structure, migrating_specie: str, max_distance: float, **kwargs
    ) -> "MigrationGraph":
        """
        Using a specific nn strategy to get the connectivity graph between all the migrating ion sites.
        Args:
            max_distance: Maximum length of NEB path in the unit
                of Angstrom. Defaults to None, which means you are setting the
                value to the min cutoff until finding 1D or >1D percolating paths.
        Returns:
            A constructed MigrationGraph object
        """
        only_sites = get_only_sites_from_structure(structure, migrating_specie)
        migration_graph = StructureGraph.with_local_env_strategy(
            only_sites,
            MinimumDistanceNN(cutoff=max_distance, get_all_sites=True),
        )
        return cls(structure=structure, migration_graph=migration_graph, **kwargs)

    @staticmethod
    def get_structure_from_entries(
        base_entries: List[ComputedStructureEntry],
        inserted_entries: List[ComputedStructureEntry],
        migrating_ion_entry: ComputedEntry,
        **kwargs,
    ) -> List[Structure]:
        """
        Read in a list of base entries and inserted entries.  Return a list of structures that contains metastable
        sites for the migration species decorated with a "insertion_energy" property.

        Args:
            base_entries: List of entries that only contains the host lattice
            inserted_entries: List of entries that contains the inserted structures
            migrating_ion_entry: The metallic phase of the working ion, used to calculate insertion energies.

        Additional Kwargs:
            symprec:  symmetry parameter for SpacegroupAnalyzer
            ltol: Fractional length tolerance for StructureMatcher
            stol: Site tolerance for StructureMatcher
            angle_tol: Angle tolerance fro StructureMatcher and SpacegroupAnalyzer
            only_single_cat: If True, only use single cation insertions so the
            site energy is more accurate use_strict_tol: halve the ltol and
            stol parameter for more strict matching.

        Returns:

        """
        l_base_and_inserted = process_entries(
            base_entries=base_entries,
            inserted_entries=inserted_entries,
            migrating_ion_entry=migrating_ion_entry,
            **kwargs,
        )
        res = []
        for group in l_base_and_inserted:
            all_sites = group["base"].copy().sites
            for isite in group["inserted"]:
                all_sites.append(isite)
            res.append(Structure.from_sites(all_sites))
        return res

    def _get_pos_and_migration_path(self, u, v, w):
        """
        insert a single MigrationHop object on a graph edge
        Args:
          u (int): index of initial node
          v (int): index of final node
          w (int): index for multiple edges that share the same two nodes
        """
        edge = self.migration_graph.graph[u][v][w]
        i_site = self.only_sites.sites[u]
        e_site = PeriodicSite(
            self.only_sites.sites[v].species,
            self.only_sites.sites[v].frac_coords + np.array(edge["to_jimage"]),
            lattice=self.only_sites.lattice,
        )
        # Positions might be useful for plotting
        edge["ipos"] = i_site.frac_coords
        edge["epos"] = e_site.frac_coords
        edge["ipos_cart"] = np.dot(i_site.frac_coords, self.only_sites.lattice.matrix)
        edge["epos_cart"] = np.dot(e_site.frac_coords, self.only_sites.lattice.matrix)

        edge["hop"] = MigrationHop(i_site, e_site, self.symm_structure)

    def _populate_edges_with_migration_paths(self):
        """
        Populate the edges with the data for the Migration Paths
        """
        list(starmap(self._get_pos_and_migration_path, self.migration_graph.graph.edges))

    def _group_and_label_hops(self):
        """
        Group the MigrationHop objects together and label all the symmetrically equlivaelnt hops with the same label
        """
        hops = list(nx.get_edge_attributes(self.migration_graph.graph, "hop").items())
        labs = generic_groupby(hops, comp=lambda x, y: x[1] == y[1])
        new_attr = {g_index: {"hop_label": labs[edge_index]} for edge_index, (g_index, _) in enumerate(hops)}
        nx.set_edge_attributes(self.migration_graph.graph, new_attr)
        return new_attr

    def add_data_to_similar_edges(
        self,
        target_label: Union[int, str],
        data: dict,
        m_path: MigrationHop = None,
    ):
        """
        Insert data to all edges with the same label
        Args:
            target_label: The edge uniqueness label are adding data
            data: The data to passed to the different edges
            m_path: If the data is an array, and m_path is set, it uses the reference migration path to
            determine whether the data needs to be flipped so that 0-->1 is different from 1-->0
        """

        for u, v, d in self.migration_graph.graph.edges(data=True):
            if d["hop_label"] == target_label:
                d.update(data)
                if m_path is not None:
                    # Try to override the data.
                    if not m_path.symm_structure.spacegroup.are_symmetrically_equivalent(
                        [m_path.isite], [d["hop"].isite]
                    ):
                        # "The data going to this edge needs to be flipped"
                        for k in data.keys():
                            if isinstance(data[k], (np.ndarray, np.generic)):
                                raise Warning("The data provided will only be flipped " "if it a list")
                            if not isinstance(data[k], list):
                                continue
                            d[k] = d[k][::-1]  # flip the data in the array

    def assign_cost_to_graph(self, cost_keys=["hop_distance"]):
        """
        Read the data dict on each add and populate a cost key
        Args:
            cost_keys: a list of keys for data on each edge.
                The SC Graph is decorated with a "cost" key that is the product of the different keys here
        """
        for k, v in self.unique_hops.items():
            cost_val = np.prod([v[ik] for ik in cost_keys])
            self.add_data_to_similar_edges(k, {"cost": cost_val})

    def get_path(self, max_val=100000):
        """
        obtain a pathway through the material using hops that are in the current graph
        Basic idea:
            Get an endpoint p1 in the graph that is outside the current unit cell
            Ask the graph for a pathway that connects to p1 from either within the (0,0,0) cell
            or any other neighboring UC not containing p1.
        Args:
            max_val: Filter the graph by a cost
        Returns:
            Generator for List of Dicts:
            Each dict contains the information of a hop
        """

        if len(self.unique_hops) != len(self.unique_hops):
            logger.error(
                f"There are {len(self.unique_hops)} SC hops but {len(self.unique_hops)} UC hops in {self.__str__()}"
            )

        # for u, v, k, d in self.migration_graph.graph.edges(data=True, keys=True):
        for u in self.migration_graph.graph.nodes():
            # Create a copy of the graph so that we can trim the higher cost hops
            path_graph = deepcopy(self.migration_graph.graph)
            # Trim the higher cost edges from the network
            cut_edges = []
            for tmp_u, tmp_v, tmp_k, tmp_d in path_graph.edges(data=True, keys=True):
                if tmp_d["cost"] > max_val:
                    cut_edges.append((tmp_u, tmp_v, tmp_k))
            for tmp_u, tmp_v, tmp_k in cut_edges:
                path_graph.remove_edge(tmp_u, tmp_v, key=tmp_k)
            # populate the entire graph with multiple images
            best_ans, path_parent = periodic_dijkstra(path_graph, sources={u}, weight="cost", max_image=2)
            # find a way to a u site that is not in the (0,0,0) image
            all_paths = []
            for idx, jimage in path_parent.keys():
                if idx == u and jimage != (0, 0, 0):
                    path = [*get_optimal_pathway_rev(path_parent, (idx, jimage))][::-1]
                    assert path[-1][0] == u
                    all_paths.append(path)

            if len(all_paths) == 0:
                continue
            # The first hop must be one that leaves the 000 unit cell
            path = min(all_paths, key=lambda x: best_ans[x[-1]])

            # get the sequence of MigrationHop objects the represent the pathway
            path_hops = []
            for (idx1, jimage1), (idx2, jimage2) in zip(path[:-1], path[1:]):
                # for each pair of points in the periodic graph path look for end points in the original graph
                # the index pair has to be u->v with u <= v
                # once that is determined look up all such pairs in the graph and see if relative image
                # displacement +/- (jimage1 - jimage2) is present on of of the edges
                # Note: there should only ever be one valid to_jimage for a u->v pair
                i1_, i2_ = sorted((idx1, idx2))
                all_edge_data = [*path_graph.get_edge_data(i1_, i2_, default={}).items()]
                image_diff = np.subtract(jimage2, jimage1)
                found_ = 0
                for k, tmp_d in all_edge_data:
                    if tmp_d["to_jimage"] in {tuple(image_diff), tuple(-image_diff)}:
                        path_hops.append(tmp_d)
                        found_ += 1
                if found_ != 1:
                    raise RuntimeError("More than one edge matched in original graph.")
            yield u, path_hops

    def get_summary_dict(self, added_keys: List[str] = None) -> dict:
        """
        Dictionary format, for saving to database
        """
        hops = []
        keys = ["hop_label", "to_jimage", "ipos", "epos", "ipos_cart", "epos_cart"]

        if added_keys is not None:
            keys += added_keys

        def get_keys(d):
            return {k_: d[k_] for k_ in keys if k_ in d}

        for u, v, d in self.migration_graph.graph.edges(data=True):
            new_hop = get_keys(d)
            new_hop["iindex"] = u
            new_hop["eindex"] = v
            hops.append(new_hop)

        unique_hops = []
        for k, d in self.unique_hops.items():
            new_hop["iindex"] = d["iindex"]
            new_hop["eindex"] = d["eindex"]
            unique_hops.append(get_keys(d))

        unique_hops = sorted(unique_hops, key=lambda x: x["hop_label"])

        return dict(
            structure=self.structure.as_dict(),
            host_structure=self.host_structure.as_dict(),
            migrating_specie=list(self.only_sites.composition.as_dict().keys()),
            hops=hops,
            unique_hops=unique_hops,
        )


class ChargeBarrierGraph(MigrationGraph):
    """
    A Migration graph with additional charge density analysis on the charge density of the host material
    """

    def __init__(
        self,
        structure: Structure,
        migration_graph: StructureGraph,
        potential_field: VolumetricData,
        potential_data_key: str,
        **kwargs,
    ):
        """
        Construct the MigrationGraph object using a VolumetricData object.
        The graph is constructure using the structure, and cost values are assigned based on charge density analysis.
        Args:
            potential_field: Input VolumetricData object that describes the field does
                not have to contains all the metastable sites.
            migrating_specie (Specie-like): The specie that migrates. E.g.,
                "Li".
            symprec (float): Symmetry precision to determine equivalence.
        """
        self.potential_field = potential_field
        self.potential_data_key = potential_data_key
        super().__init__(structure=structure, migration_graph=migration_graph, **kwargs)
        self._setup_grids()

    def _setup_grids(self):
        """Populate the internal varialbes used for defining the grid points in the charge density analysis"""

        # set up the grid
        aa = np.linspace(0, 1, len(self.potential_field.get_axis_grid(0)), endpoint=False)
        bb = np.linspace(0, 1, len(self.potential_field.get_axis_grid(1)), endpoint=False)
        cc = np.linspace(0, 1, len(self.potential_field.get_axis_grid(2)), endpoint=False)
        # move the grid points to the center
        aa, bb, dd = map(_shift_grid, [aa, bb, cc])

        # mesh grid for each unit cell
        AA, BB, CC = np.meshgrid(aa, bb, cc, indexing="ij")

        # should be using a mesh grid of 5x5x5 (using 3x3x3 misses some fringe cases)
        # but using 3x3x3 is much faster and only crops the cyliners in some rare case
        # if you keep the tube_radius small then this is not a big deal
        IMA, IMB, IMC = np.meshgrid([-1, 0, 1], [-1, 0, 1], [-1, 0, 1], indexing="ij")

        # store these
        self._uc_grid_shape = AA.shape
        self._fcoords = np.vstack([AA.flatten(), BB.flatten(), CC.flatten()]).T
        self._images = np.vstack([IMA.flatten(), IMB.flatten(), IMC.flatten()]).T

    def _dist_mat(self, pos_frac):
        # return a matrix that contains the distances to pos_frac
        aa = np.linspace(0, 1, len(self.potential_field.get_axis_grid(0)), endpoint=False)
        bb = np.linspace(0, 1, len(self.potential_field.get_axis_grid(1)), endpoint=False)
        cc = np.linspace(0, 1, len(self.potential_field.get_axis_grid(2)), endpoint=False)
        aa, bb, cc = map(_shift_grid, [aa, bb, cc])
        AA, BB, CC = np.meshgrid(aa, bb, cc, indexing="ij")
        dist_from_pos = self.potential_field.structure.lattice.get_all_distances(
            fcoords1=np.vstack([AA.flatten(), BB.flatten(), CC.flatten()]).T,
            fcoords2=pos_frac,
        )
        return dist_from_pos.reshape(AA.shape)

    def _get_pathfinder_from_hop(self, migration_path: MigrationHop, n_images=20):
        # get migration pathfinder objects which contains the paths
        ipos = migration_path.isite.frac_coords
        epos = migration_path.esite.frac_coords
        mpos = migration_path.esite.frac_coords

        start_struct = self.potential_field.structure.copy()
        end_struct = self.potential_field.structure.copy()
        mid_struct = self.potential_field.structure.copy()

        # the moving ion is always inserted on the zero index
        start_struct.insert(0, migration_path.isite.species_string, ipos, properties=dict(magmom=0))
        end_struct.insert(0, migration_path.isite.species_string, epos, properties=dict(magmom=0))
        mid_struct.insert(0, migration_path.isite.species_string, mpos, properties=dict(magmom=0))

        chgpot = ChgcarPotential(self.potential_field, normalize=False)
        npf = NEBPathfinder(
            start_struct,
            end_struct,
            relax_sites=[0],
            v=chgpot.get_v(),
            n_images=n_images,
            mid_struct=mid_struct,
        )
        return npf

    def _get_avg_chg_at_max(self, migration_path, radius=None, chg_along_path=False, output_positions=False):
        """obtain the maximum average charge along the path
        Args:
            migration_path (MigrationHop): MigrationPath object that represents a given hop
            radius (float, optional): radius of sphere to perform the average.
                    Defaults to None, which used the _tube_radius instead
            chg_along_path (bool, optional): If True, also return the entire list of average
                    charges along the path for plotting.
                    Defaults to False.
            output_positions (bool, optional): If True, also return the entire list of average
                    charges along the path for plotting.
                    Defaults to False.
        Returns:
            [float]: maximum of the charge density, (optional: entire list of charge density)
        """
        if radius is None:
            rr = self._tube_radius
        if rr <= 0:
            raise ValueError("The integration radius must be positive.")

        npf = self._get_pathfinder_from_hop(migration_path)
        # get the charge in a sphere around each point
        centers = [image.sites[0].frac_coords for image in npf.images]
        avg_chg = []
        for ict in centers:
            dist_mat = self._dist_mat(ict)
            mask = dist_mat < rr
            vol_sphere = self.potential_field.structure.volume * (mask.sum() / self.potential_field.ngridpts)
            avg_chg.append(
                np.sum(self.potential_field.data[self.potential_data_key] * mask)
                / self.potential_field.ngridpts
                / vol_sphere
            )
        if output_positions:
            return max(avg_chg), avg_chg, centers
        if chg_along_path:
            return max(avg_chg), avg_chg
        return max(avg_chg)

    def _get_chg_between_sites_tube(self, migration_path, mask_file_seedname=None):
        """
        Calculate the amount of charge that a migrating ion has to move through in order to complete a hop
        Args:
            migration_path: MigrationHop object that represents a given hop
            mask_file_seedname(string): seedname for output of the migration path masks (for debugging and
                visualization) (Default value = None)
        Returns:
            float: The total charge density in a tube that connects two sites of a given edges of the graph
        """
        try:
            self._tube_radius
        except NameError:
            logger.warning("The radius of the tubes for charge analysis need to be defined first.")
        ipos = migration_path.isite.frac_coords
        epos = migration_path.esite.frac_coords

        cart_ipos = np.dot(ipos, self.potential_field.structure.lattice.matrix)
        cart_epos = np.dot(epos, self.potential_field.structure.lattice.matrix)
        pbc_mask = np.zeros(self._uc_grid_shape, dtype=bool).flatten()
        for img in self._images:
            grid_pos = np.dot(self._fcoords + img, self.potential_field.structure.lattice.matrix)
            proj_on_line = np.dot(grid_pos - cart_ipos, cart_epos - cart_ipos) / (np.linalg.norm(cart_epos - cart_ipos))
            dist_to_line = np.linalg.norm(
                np.cross(grid_pos - cart_ipos, cart_epos - cart_ipos) / (np.linalg.norm(cart_epos - cart_ipos)),
                axis=-1,
            )

            mask = (
                (proj_on_line >= 0)
                * (proj_on_line < np.linalg.norm(cart_epos - cart_ipos))
                * (dist_to_line < self._tube_radius)
            )
            pbc_mask = pbc_mask + mask
        pbc_mask = pbc_mask.reshape(self._uc_grid_shape)

        if mask_file_seedname:
            mask_out = VolumetricData(
                structure=self.potential_field.structure.copy(),
                data={"total": self.potential_field.data["total"]},
            )
            mask_out.structure.insert(0, "X", ipos)
            mask_out.structure.insert(0, "X", epos)
            mask_out.data[self.potential_data_key] = pbc_mask
            isym = self.symm_structure.wyckoff_symbols[migration_path.iindex]
            esym = self.symm_structure.wyckoff_symbols[migration_path.eindex]
            mask_out.write_file(
                "{}_{}_{}_tot({:0.2f}).vasp".format(
                    mask_file_seedname,
                    isym,
                    esym,
                    mask_out.data[self.potential_data_key].sum(),
                )
            )

        return (
            self.potential_field.data[self.potential_data_key][pbc_mask].sum()
            / self.potential_field.ngridpts
            / self.potential_field.structure.volume
        )

    def populate_edges_with_chg_density_info(self, tube_radius=1):
        """
        Args:
            tube_radius: Tube radius.
        """
        self._tube_radius = tube_radius
        for k, v in self.unique_hops.items():
            # charge in tube
            chg_tot = self._get_chg_between_sites_tube(v["hop"])
            self.add_data_to_similar_edges(k, {"chg_total": chg_tot})

            # max charge in sphere
            max_chg, avg_chg_list, frac_coords_list = self._get_avg_chg_at_max(
                v["hop"], chg_along_path=True, output_positions=True
            )
            images = [
                {"position": ifrac, "average_charge": ichg} for ifrac, ichg in zip(frac_coords_list, avg_chg_list)
            ]
            v.update(
                dict(
                    chg_total=chg_tot,
                    max_avg_chg=max_chg,
                    images=images,
                )
            )
            self.add_data_to_similar_edges(k, {"max_avg_chg": max_chg})

    def get_least_chg_path(self):
        """
        obtain an intercolating pathway through the material that has the least amount of charge
        Returns:
            list of hops
        """
        min_chg = 100000000
        min_path = []
        all_paths = self.get_path()
        for path in all_paths:
            sum_chg = np.sum([hop[2]["chg_total"] for hop in path])
            sum_length = np.sum([hop[2]["hop"].length for hop in path])
            avg_chg = sum_chg / sum_length
            if avg_chg < min_chg:
                min_chg = sum_chg
                min_path = path
        return min_path

    def get_summary_dict(self, add_keys: List[str] = None):
        """
        Dictionary format, for saving to database
        """
        a_keys = ["max_avg_chg", "chg_total"]
        if add_keys is not None:
            a_keys += add_keys
        return super().get_summary_dict(added_keys=a_keys)


# Utility functions
def get_only_sites_from_structure(structure: Structure, migrating_specie: str) -> Structure:
    """
    Get a copy of the structure with only the migrating sites.
    Args:
        structure: The full_structure that contains all the sites
        migrating_specie: The name of migrating species
    Returns:
      Structure: Structure with all possible migrating ion sites
    """
    migrating_ion_sites = list(
        filter(
            lambda site: site.species == Composition({migrating_specie: 1}),
            structure.sites,
        )
    )
    return Structure.from_sites(migrating_ion_sites)


def _shift_grid(vv):
    """
    Move the grid points by half a step so that they sit in the center
    Args:
        vv: equally space grid points in 1-D
    """
    step = vv[1] - vv[0]
    return vv + step / 2.0


def get_hop_site_sequence(hop_list: List[Dict], start_u: Union[int, str]) -> List:
    """
    Read in a list of hop dictionaries and print the sequence of sites.
    Args:
        hop_list: a list of the data on a sequence of hops
        start_u: the site index of the starting sites
    Returns:
        String representation of the hop sequence
    """
    hops = iter(hop_list)
    ihop = next(hops)
    if ihop["eindex"] == start_u:
        site_seq = [ihop["eindex"], ihop["iindex"]]
    else:
        site_seq = [ihop["iindex"], ihop["eindex"]]

    for ihop in hops:
        if ihop["iindex"] == site_seq[-1]:
            site_seq.append(ihop["eindex"])
        elif ihop["eindex"] == site_seq[-1]:
            site_seq.append(ihop["iindex"])
        else:
            raise RuntimeError("The sequence of sites for the path is invalid.")
    return site_seq


"""
Note the current pathway algorithm no longer needs supercells but the following
functions might still be useful for other applications
Finding all possible pathways in the periodic network is not possible.
We can do a good enough job if we make a (2x2x2) supercell of the structure and find
migration events using the following procedure:
- Look for a hop that leaves the SC like A->B (B on the outside)
- then at A look for a pathway to the image of B inside the SC
"""


# Utility Functions for comparing UC and SC hops


def almost(a, b):
    """
    return true if the values are almost equal
    """
    SMALL_VAL = 1e-4
    try:
        return all(almost(i, j) for i, j in zip(list(a), list(b)))
    except BaseException:
        if isinstance(a, (int, float)) and isinstance(b, (int, float)):
            return abs(a - b) < SMALL_VAL
        raise NotImplementedError


def check_uc_hop(sc_hop, uc_hop):
    """
    See if hop in the 2X2X2 supercell and a unit cell hop
    are equilvalent under lattice translation
    Args:
        sc_hop: MigrationHop object form pymatgen-diffusion.
        uc_hop: MigrationHop object form pymatgen-diffusion.
    Return:
        image vector of lenght 3
        Is the UC hop flip of the SC hop
    """

    directions = np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1],
            [0, 1, 1],
            [1, 0, 1],
            [1, 1, 0],
            [1, 1, 1],
        ]
    )

    sc_ipos = [icoord * 2 for icoord in sc_hop.isite.frac_coords]
    sc_epos = [icoord * 2 for icoord in sc_hop.esite.frac_coords]
    sc_mpos = [icoord * 2 for icoord in sc_hop.msite.frac_coords]

    uc_ipos = uc_hop.isite.frac_coords
    uc_epos = uc_hop.esite.frac_coords
    uc_mpos = uc_hop.msite.frac_coords

    for idir in directions:
        tmp_m = uc_mpos + idir
        if almost(tmp_m, sc_mpos):
            tmp_i = uc_ipos + idir
            tmp_e = uc_epos + idir
            if almost(tmp_i, sc_ipos) and almost(tmp_e, sc_epos):
                return idir, False
            if almost(tmp_e, sc_ipos) and almost(tmp_i, sc_epos):
                return idir, True
    return None


def map_hop_sc2uc(sc_hop, fpm_uc):
    """
    Map a given hop in the SC onto the UC.
    Args:
        sc_hop: MigrationHop object form pymatgen-diffusion.
        fpm_uc: MigrationGraph object from pymatgen-diffusion.
    Note:
        For now assume that the SC is exactly 2x2x2 of the UC.
        Can add in the parsing of different SC's later
        For a migration event in the SC from (0.45,0,0)-->(0.55,0,0)
        the UC hop might be (0.9,0,0)-->(0.1,0,0)[img:100]
        for the inverse of (0.1,0,0)-->(-0.1,0,0) the code needs to account for both those cases
    """
    for u, v, d in fpm_uc.migration_graph.graph.edges(data=True):
        chk_res = check_uc_hop(sc_hop=sc_hop, uc_hop=d["hop"])
        if chk_res is not None:
            assert almost(d["hop"].length, sc_hop.length)
            return dict(
                uc_u=u,
                uc_v=v,
                hop=d["hop"],
                shift=chk_res[0],
                flip=chk_res[1],
                hop_label=d["hop_label"],
            )
    raise AssertionError("Looking for a SC hop without a matching UC hop")



# coding: utf-8
# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.

"""
Algorithms for NEB migration path analysis.
"""

import itertools
import logging
import warnings
from typing import Tuple

import numpy as np
from pymatgen.core import Site
from pymatgen.core import PeriodicSite, Structure
from pymatgen.core.periodic_table import get_el_sp
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

__author__ = "Iek-Heng Chu"
__version__ = "1.0"
__date__ = "March 14, 2017"


# TODO: (1) ipython notebook example files, unittests
from pymatgen.symmetry.structure import SymmetrizedStructure

from pymatgen.analysis.diffusion.utils.supercells import (
    get_sc_fromstruct,
    get_start_end_structures,
)

logger = logging.getLogger(__name__)


class IDPPSolver:
    """
    A solver using image dependent pair potential (IDPP) algo to get an improved
    initial NEB path. For more details about this algo, please refer to
    Smidstrup et al., J. Chem. Phys. 140, 214106 (2014).
    """

    def __init__(self, structures):
        """
        Initialization.

        Args:
            structures (list of pmg_structure) : Initial guess of the NEB path
                (including initial and final end-point structures).
        """

        latt = structures[0].lattice
        natoms = structures[0].num_sites
        nimages = len(structures) - 2
        target_dists = []

        # Initial guess of the path (in Cartesian coordinates) used in the IDPP
        # algo.
        init_coords = []

        # Construct the set of target distance matrices via linear interpolation
        # between those of end-point structures.
        for i in range(1, nimages + 1):
            # Interpolated distance matrices
            dist = structures[0].distance_matrix + i / (nimages + 1) * (
                structures[-1].distance_matrix - structures[0].distance_matrix
            )

            target_dists.append(dist)

        target_dists = np.array(target_dists)

        # Set of translational vector matrices (anti-symmetric) for the images.
        translations = np.zeros((nimages, natoms, natoms, 3), dtype=np.float64)

        # A set of weight functions. It is set as 1/d^4 for each image. Here,
        # we take d as the average of the target distance matrix and the actual
        # distance matrix.
        weights = np.zeros_like(target_dists, dtype=np.float64)
        for ni in range(nimages):
            avg_dist = (target_dists[ni] + structures[ni + 1].distance_matrix) / 2.0
            weights[ni] = 1.0 / (avg_dist ** 4 + np.eye(natoms, dtype=np.float64) * 1e-8)

        for ni, i in itertools.product(range(nimages + 2), range(natoms)):
            frac_coords = structures[ni][i].frac_coords
            init_coords.append(latt.get_cartesian_coords(frac_coords))

            if ni not in [0, nimages + 1]:
                for j in range(i + 1, natoms):
                    img = latt.get_distance_and_image(frac_coords, structures[ni][j].frac_coords)[1]
                    translations[ni - 1, i, j] = latt.get_cartesian_coords(img)
                    translations[ni - 1, j, i] = -latt.get_cartesian_coords(img)

        self.init_coords = np.array(init_coords).reshape(nimages + 2, natoms, 3)
        self.translations = translations
        self.weights = weights
        self.structures = structures
        self.target_dists = target_dists
        self.nimages = nimages

    def run(
        self,
        maxiter=1000,
        tol=1e-5,
        gtol=1e-3,
        step_size=0.05,
        max_disp=0.05,
        spring_const=5.0,
        species=None,
    ):
        """
        Perform iterative minimization of the set of objective functions in an
        NEB-like manner. In each iteration, the total force matrix for each
        image is constructed, which comprises both the spring forces and true
        forces. For more details about the NEB approach, please see the
        references, e.g. Henkelman et al., J. Chem. Phys. 113, 9901 (2000).

        Args:
            maxiter (int): Maximum number of iterations in the minimization
                process.
            tol (float): Tolerance of the change of objective functions between
                consecutive steps.
            gtol (float): Tolerance of maximum force component (absolute value).
            step_size (float): Step size associated with the displacement of
                the atoms during the minimization process.
            max_disp (float): Maximum allowed atomic displacement in each
                iteration.
            spring_const (float): A virtual spring constant used in the NEB-like
                        relaxation process that yields so-called IDPP path.
            species (list of string): If provided, only those given species are
                allowed to move. The atomic positions of other species are
                obtained via regular linear interpolation approach.

        Returns:
            [Structure] Complete IDPP path (including end-point structures)
        """

        coords = self.init_coords.copy()
        old_funcs = np.zeros((self.nimages,), dtype=np.float64)
        idpp_structures = [self.structures[0]]

        if species is None:
            indices = list(range(len(self.structures[0])))
        else:
            species = [get_el_sp(sp) for sp in species]
            indices = [i for i, site in enumerate(self.structures[0]) if site.specie in species]

            if len(indices) == 0:
                raise ValueError("The given species are not in the system!")

        # Iterative minimization
        for n in range(maxiter):
            # Get the sets of objective functions, true and total force
            # matrices.
            funcs, true_forces = self._get_funcs_and_forces(coords)
            tot_forces = self._get_total_forces(coords, true_forces, spring_const=spring_const)

            # Each atom is allowed to move up to max_disp
            disp_mat = step_size * tot_forces[:, indices, :]
            disp_mat = np.where(np.abs(disp_mat) > max_disp, np.sign(disp_mat) * max_disp, disp_mat)
            coords[1 : (self.nimages + 1), indices] += disp_mat

            max_force = np.abs(tot_forces[:, indices, :]).max()
            tot_res = np.sum(np.abs(old_funcs - funcs))

            if tot_res < tol and max_force < gtol:
                break

            old_funcs = funcs

        else:
            warnings.warn("Maximum iteration number is reached without convergence!", UserWarning)

        for ni in range(self.nimages):
            # generate the improved image structure
            new_sites = []

            for site, cart_coords in zip(self.structures[ni + 1], coords[ni + 1]):
                new_site = PeriodicSite(
                    site.species,
                    coords=cart_coords,
                    lattice=site.lattice,
                    coords_are_cartesian=True,
                    properties=site.properties,
                )
                new_sites.append(new_site)

            idpp_structures.append(Structure.from_sites(new_sites))

        # Also include end-point structure.
        idpp_structures.append(self.structures[-1])

        return idpp_structures

    @classmethod
    def from_endpoints(cls, endpoints, nimages=5, sort_tol=1.0):
        """
        A class method that starts with end-point structures instead. The
        initial guess for the IDPP algo is then constructed using linear
        interpolation.

        Args:
            endpoints (list of Structure objects): The two end-point structures.
            nimages (int): Number of images between the two end-points.
            sort_tol (float): Distance tolerance (in Angstrom) used to match the
                atomic indices between start and end structures. Need
                to increase the value in some cases.
        """
        try:
            images = endpoints[0].interpolate(endpoints[1], nimages=nimages + 1, autosort_tol=sort_tol)
        except Exception as e:
            if "Unable to reliably match structures " in str(e):
                warnings.warn(
                    "Auto sorting is turned off because it is unable" " to match the end-point structures!",
                    UserWarning,
                )
                images = endpoints[0].interpolate(endpoints[1], nimages=nimages + 1, autosort_tol=0)
            else:
                raise e

        return IDPPSolver(images)

    def _get_funcs_and_forces(self, x):
        """
        Calculate the set of objective functions as well as their gradients,
        i.e. "effective true forces"
        """
        funcs = []
        funcs_prime = []
        trans = self.translations
        natoms = trans.shape[1]
        weights = self.weights
        target_dists = self.target_dists

        for ni in range(len(x) - 2):
            vec = [x[ni + 1, i] - x[ni + 1] - trans[ni, i] for i in range(natoms)]

            trial_dist = np.linalg.norm(vec, axis=2)
            aux = (trial_dist - target_dists[ni]) * weights[ni] / (trial_dist + np.eye(natoms, dtype=np.float64))

            # Objective function
            func = np.sum((trial_dist - target_dists[ni]) ** 2 * weights[ni])

            # "True force" derived from the objective function.
            grad = np.sum(aux[:, :, None] * vec, axis=1)

            funcs.append(func)
            funcs_prime.append(grad)

        return 0.5 * np.array(funcs), -2 * np.array(funcs_prime)

    @staticmethod
    def get_unit_vector(vec):
        """
        Calculate the unit vector of a vector.
        Args:
            vec: Vector.
        """
        return vec / np.sqrt(np.sum(vec ** 2))

    def _get_total_forces(self, x, true_forces, spring_const):
        """
        Calculate the total force on each image structure, which is equal to
        the spring force along the tangent + true force perpendicular to the
        tangent. Note that the spring force is the modified version in the
        literature (e.g. Henkelman et al., J. Chem. Phys. 113, 9901 (2000)).
        """

        total_forces = []
        natoms = np.shape(true_forces)[1]

        for ni in range(1, len(x) - 1):
            vec1 = (x[ni + 1] - x[ni]).flatten()
            vec2 = (x[ni] - x[ni - 1]).flatten()

            # Local tangent
            tangent = self.get_unit_vector(vec1) + self.get_unit_vector(vec2)
            tangent = self.get_unit_vector(tangent)

            # Spring force
            spring_force = spring_const * (np.linalg.norm(vec1) - np.linalg.norm(vec2)) * tangent

            # Total force
            flat_ft = true_forces[ni - 1].copy().flatten()
            total_force = true_forces[ni - 1] + (spring_force - np.dot(flat_ft, tangent) * tangent).reshape(natoms, 3)
            total_forces.append(total_force)

        return np.array(total_forces)


class MigrationHop:
    """
    A convenience container representing a migration path.
    """

    def __init__(
        self,
        isite: Site,
        esite: Site,
        symm_structure: SymmetrizedStructure,
        symprec: float = 0.001,
    ):
        """
        Args:
            isite: Initial site
            esite: End site
            symm_structure: SymmetrizedStructure
            symprec: used to determine equivalence
        """
        self.isite = isite
        self.esite = esite
        self.iindex = None
        self.eindex = None
        self.symm_structure = symm_structure
        self.symprec = symprec
        self.msite = PeriodicSite(esite.specie, (isite.frac_coords + esite.frac_coords) / 2, esite.lattice)
        sg = self.symm_structure.spacegroup
        for i, sites in enumerate(self.symm_structure.equivalent_sites):
            if sg.are_symmetrically_equivalent([isite], [sites[0]]):
                self.iindex = i
            if sg.are_symmetrically_equivalent([esite], [sites[0]]):
                self.eindex = i

        # if no index was identified then loop over each site until something is found
        if self.iindex is None:
            for i, sites in enumerate(self.symm_structure.equivalent_sites):
                for itr_site in sites:
                    if sg.are_symmetrically_equivalent([isite], [itr_site]):
                        self.iindex = i
                        break
                else:
                    continue
                break
        if self.eindex is None:
            for i, sites in enumerate(self.symm_structure.equivalent_sites):
                for itr_site in sites:
                    if sg.are_symmetrically_equivalent([esite], [itr_site]):
                        self.eindex = i
                        break
                else:
                    continue
                break

        if self.iindex is None:
            raise RuntimeError(f"No symmetrically equivalent site was found for {isite}")
        if self.eindex is None:
            raise RuntimeError(f"No symmetrically equivalent site was found for {esite}")

    def __repr__(self):
        return (
            "Path of %.4f A from %s [%.3f, %.3f, %.3f] "
            "(ind: %d, Wyckoff: %s) to %s [%.3f, %.3f, %.3f] (ind: %d, Wyckoff: %s)"
            % (
                self.length,
                self.isite.specie,
                self.isite.frac_coords[0],
                self.isite.frac_coords[1],
                self.isite.frac_coords[2],
                self.iindex,
                self.symm_structure.wyckoff_symbols[self.iindex],
                self.esite.specie,
                self.esite.frac_coords[0],
                self.esite.frac_coords[1],
                self.esite.frac_coords[2],
                self.eindex,
                self.symm_structure.wyckoff_symbols[self.eindex],
            )
        )

    @property
    def length(self):
        """
        Returns:
            (float) Length of migration path.
        """
        return np.linalg.norm(self.isite.coords - self.esite.coords)

    def __hash__(self):
        return self.iindex + self.eindex

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if self.symm_structure != other.symm_structure:
            return False

        if abs(self.length - other.length) > 1e-3:
            return False

        return self.symm_structure.spacegroup.are_symmetrically_equivalent(
            (self.isite, self.msite, self.esite),
            (other.isite, other.msite, other.esite),
            self.symprec,
        )

    def get_structures(self, nimages=5, vac_mode=True, idpp=False, **idpp_kwargs):
        r"""
        Generate structures for NEB calculation.

        Args:
            nimages (int): Defaults to 5. Number of NEB images. Total number of
                structures returned in nimages+2.
            vac_mode (bool): Defaults to True. In vac_mode, a vacancy diffusion
                mechanism is assumed. The initial and end sites of the path
                are assumed to be the initial and ending positions of the
                vacancies. If vac_mode is False, an interstitial mechanism is
                assumed. The initial and ending positions are assumed to be
                the initial and ending positions of the interstitial, and all
                other sites of the same specie are removed. E.g., if NEBPaths
                were obtained using a Li4Fe4P4O16 structure, vac_mode=True would
                generate structures with formula Li3Fe4P4O16, while
                vac_mode=False would generate structures with formula
                LiFe4P4O16.
            idpp (bool): Defaults to False. If True, the generated structures
                will be run through the IDPPSolver to generate a better guess
                for the minimum energy path.
            \*\*idpp_kwargs: Passthrough kwargs for the IDPPSolver.run.

        Returns:
            [Structure] Note that the first site of each structure is always
            the migrating ion. This makes it easier to perform subsequent
            analysis.
        """
        migrating_specie_sites, other_sites = self._split_migrating_and_other_sites(vac_mode)

        start_structure = Structure.from_sites([self.isite] + migrating_specie_sites + other_sites)
        end_structure = Structure.from_sites([self.esite] + migrating_specie_sites + other_sites)

        structures = start_structure.interpolate(end_structure, nimages=nimages + 1, pbc=False)

        if idpp:
            solver = IDPPSolver(structures)
            return solver.run(**idpp_kwargs)

        return structures

    def _split_migrating_and_other_sites(self, vac_mode):
        migrating_specie_sites = []
        other_sites = []
        for site in self.symm_structure.sites:
            if site.specie != self.isite.specie:
                other_sites.append(site)
            else:
                if not vac_mode or (self.isite.distance(site) <= 1e-8 or self.esite.distance(site) <= 1e-8):
                    continue
                migrating_specie_sites.append(site)
        return migrating_specie_sites, other_sites

    def get_sc_structures(
        self,
        vac_mode: bool,
        min_atoms: int = 80,
        max_atoms: int = 240,
        min_length: float = 10.0,
    ) -> Tuple[Structure, Structure, Structure]:
        """
        Construct supercells that represents the start and end positions for migration analysis.

        Args:
            vac_mode: If true simulate vacancy diffusion.
            max_atoms: Maximum number of atoms allowed in the supercell.
            min_atoms: Minimum number of atoms allowed in the supercell.
            min_length: Minimum length of the smallest supercell lattice vector.

        Returns:
            Start, End, Base Structures.

            If not vacancy mode, the base structure is just the host lattice.
            If in vacancy mode, the base structure is the fully intercalated structure

        """
        migrating_specie_sites, other_sites = self._split_migrating_and_other_sites(vac_mode)
        base_struct = Structure.from_sites(other_sites)
        sc_mat = get_sc_fromstruct(
            base_struct=base_struct,
            min_atoms=min_atoms,
            max_atoms=max_atoms,
            min_length=min_length,
        )
        start_struct, end_struct, base_sc = get_start_end_structures(
            self.isite, self.esite, base_struct, sc_mat, vac_mode=vac_mode  # type: ignore
        )
        return start_struct, end_struct, base_sc

    def write_path(self, fname, **kwargs):
        r"""
        Write the path to a file for easy viewing.

        Args:
            fname (str): File name.
            \*\*kwargs: Kwargs supported by NEBPath.get_structures.
        """
        sites = []
        for st in self.get_structures(**kwargs):
            sites.extend(st)
        st = Structure.from_sites(sites)
        st.to(filename=fname)


class DistinctPathFinder:
    """
    Determines symmetrically distinct paths between existing sites.
    The path info can then be used to set up either vacancy or interstitial
    diffusion (assuming site positions are known). Note that this works mainly
    for atomic mechanism, and does not work for correlated migration.
    """

    def __init__(
        self,
        structure,
        migrating_specie,
        max_path_length=None,
        symprec=0.1,
        perc_mode=">1d",
    ):
        """
        Args:
            structure: Input structure that contains all sites.
            migrating_specie (Specie-like): The specie that migrates. E.g.,
                "Li".
            max_path_length (float): Maximum length of NEB path in the unit
                of Angstrom. Defaults to None, which means you are setting the
                value to the min cutoff until finding 1D or >1D percolating paths.
            symprec (float): Symmetry precision to determine equivalence.
            perc_mode(str): The percolating type. Default to ">1d", because usually
                it is used to find possible NEB paths to form percolating networks.
                If you just want to check the min 1D percolation, set it to "1d".
        """
        self.structure = structure
        self.migrating_specie = get_el_sp(migrating_specie)
        self.symprec = symprec
        a = SpacegroupAnalyzer(self.structure, symprec=self.symprec)
        self.symm_structure = a.get_symmetrized_structure()

        junc = 0
        distance_list = []
        max_r = max_path_length or min(structure.lattice.abc)
        junc_cutoff = max_r
        for sites in self.symm_structure.equivalent_sites:
            if sites[0].specie == self.migrating_specie:
                site0 = sites[0]
                dists = []
                neighbors = self.symm_structure.get_neighbors(site0, r=max_r)
                for nn in sorted(neighbors, key=lambda nn: nn.nn_distance):
                    if nn.specie == self.migrating_specie:
                        dists.append(nn.nn_distance)
                if len(dists) > 2:
                    junc += 1
                distance_list.append(dists)
        # Avoid isolated atoms (# of neighbors < 2)
        if len(sorted(distance_list, key=len)[0]) < 2:
            path_cutoff = max_r
        # We define junction as atoms have at least three paths including
        # equivalent ones.
        elif junc == 0:
            path_cutoff = sorted(distance_list, key=lambda d: d[1])[-1][1]
        else:
            # distance_list are sorted as [[a0,a1,a2],[b0,b1,b2],[c0,c1,c2],...]
            # in which a0<a1<a2,b0<b1<b2,...
            # path_cutoff = max(a1,b1,c1,...), junc_cutoff=min(a2,b2,c2)
            path_cutoff = sorted(distance_list, key=lambda d: d[1])[-1][1]
            junc_distance_list = [d for d in distance_list if len(d) >= 3]
            junc_cutoff = sorted(junc_distance_list, key=lambda d: d[2])[0][2]

        if max_path_length is None:
            if perc_mode.lower() == "1d":
                self.max_path_length = path_cutoff
            else:
                self.max_path_length = max(junc_cutoff, path_cutoff)
        else:
            self.max_path_length = max_path_length

    def get_paths(self):
        """
        Returns:
            [MigrationHop] All distinct migration paths.
        """
        paths = set()
        for sites in self.symm_structure.equivalent_sites:
            if sites[0].specie == self.migrating_specie:
                site0 = sites[0]
                for nn in self.symm_structure.get_neighbors(site0, r=round(self.max_path_length, 3) + 0.01):
                    if nn.specie == self.migrating_specie:
                        path = MigrationHop(site0, nn, self.symm_structure)
                        paths.add(path)

        return sorted(paths, key=lambda p: p.length)

    def write_all_paths(self, fname, nimages=5, **kwargs):
        r"""
        Write a file containing all paths, using hydrogen as a placeholder for
        the images. H is chosen as it is the smallest atom. This is extremely
        useful for path visualization in a standard software like VESTA.

        Args:
            fname (str): Filename
            nimages (int): Number of images per path.
            \*\*kwargs: Passthrough kwargs to path.get_structures.
        """
        sites = []
        for p in self.get_paths():
            structures = p.get_structures(nimages=nimages, species=[self.migrating_specie], **kwargs)
            sites.append(structures[0][0])
            sites.append(structures[-1][0])
            for s in structures[1:-1]:
                sites.append(PeriodicSite("H", s[0].frac_coords, s.lattice))
        sites.extend(structures[0].sites[1:])
        Structure.from_sites(sites).to(filename=fname)



# coding: utf-8
# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.

"""
Generate input fiels for NEB calculations.
"""

import copy

from pymatgen.io.vasp.sets import MITRelaxSet, MITNEBSet
from pymatgen.core import Structure

__author__ = "Austen"


class MVLCINEBEndPointSet(MITRelaxSet):
    """
    Class for writing NEB end points relaxation inputs.
    """

    def __init__(self, structure, **kwargs):
        r"""
        Args:
            structure: Structure
            \*\*kwargs: Keyword args supported by VaspInputSets.
        """
        user_incar_settings = kwargs.get("user_incar_settings", {})
        defaults = {
            "ISIF": 2,
            "EDIFF": 5e-5,
            "EDIFFG": -0.02,
            "ISMEAR": 0,
            "ISYM": 0,
            "LCHARG": False,
            "LDAU": False,
            "NELMIN": 4,
        }

        if user_incar_settings != {}:
            defaults.update(user_incar_settings)
        kwargs["user_incar_settings"] = defaults

        super().__init__(structure, **kwargs)


class MVLCINEBSet(MITNEBSet):
    """
    MAVRL-tested settings for CI-NEB calculations. Note that these parameters
    requires the VTST modification of VASP from the Henkelman group. See
    http://theory.cm.utexas.edu/vtsttools/
    """

    def __init__(self, structures, **kwargs):
        r"""
        Args:
            structure: Structure
            \*\*kwargs: Keyword args supported by VaspInputSets.
        """
        user_incar_settings = kwargs.get("user_incar_settings", {})

        # CI-NEB settings
        defaults = {
            "EDIFF": 5e-5,
            "EDIFFG": -0.02,
            "IBRION": 3,
            "ICHAIN": 0,
            "IOPT": 1,
            "ISIF": 2,
            "ISMEAR": 0,
            "ISPIN": 2,
            "LCHARG": False,
            "LCLIMB": True,
            "LDAU": False,
            "LORBIT": 0,
            "NSW": 200,
            "POTIM": 0,
            "SPRING": -5,
        }
        if user_incar_settings != {}:
            defaults.update(user_incar_settings)

        kwargs["user_incar_settings"] = defaults

        super().__init__(structures, **kwargs)


def get_endpoints_from_index(structure, site_indices):
    """
    This class reads in one perfect structure and the two endpoint structures
    are generated using site_indices.

    Args:
        structure (Structure): A perfect structure.
        site_indices (list of int): a two-element list indicating site indices.

    Returns:
        endpoints (list of Structure): a two-element list of two endpoints
                                        Structure object.
    """

    if len(site_indices) != 2 or len(set(site_indices)) != 2:
        raise ValueError("Invalid indices!")
    if structure[site_indices[0]].specie != structure[site_indices[1]].specie:
        raise ValueError("The site indices must be " "associated with identical species!")

    s = structure.copy()
    sites = s.sites

    # Move hopping atoms to the beginning of species index.
    init_site = sites[site_indices[0]]
    final_site = sites[site_indices[1]]
    sites.remove(init_site)
    sites.remove(final_site)

    init_sites = copy.deepcopy(sites)
    final_sites = copy.deepcopy(sites)

    init_sites.insert(0, final_site)
    final_sites.insert(0, init_site)

    s_0 = Structure.from_sites(init_sites)
    s_1 = Structure.from_sites(final_sites)

    endpoints = [s_0, s_1]

    return endpoints


def get_endpoint_dist(ep_0, ep_1):
    """
    Calculate a list of site distances between two endpoints, assuming periodic
    boundary conditions.
    Args:
        ep_0 (Structure): the first endpoint structure.
        ep_1 (Structure): the second endpoint structure.
    Returns:
        dist (list): a list of distances between two structures.
    """
    ep_0.remove_oxidation_states()
    ep_1.remove_oxidation_states()
    assert ep_0.species == ep_1.species, "Formula mismatch!"
    assert ep_0.lattice.abc == ep_0.lattice.abc, "Lattice mismatch!"

    distances = []
    for site0, site1 in zip(ep_0, ep_1):
        fc = (site0.frac_coords, site1.frac_coords)
        d = ep_0.lattice.get_distance_and_image(fc[0], fc[1])[0]
        distances.append(d)

    return distances
